#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Laurent El Shafey <Laurent.El-Shafey@idiap.ch>
# Roy Wallace <roy.wallace@idiap.ch>
# isort: skip_file


from .resources import *  # noqa: F401,F403
from .io import *  # noqa: F401,F403
import inspect
import numpy


def score_fusion_strategy(strategy_name="average"):
    """Returns a function to compute a fusion strategy between different scores.

    Different strategies are employed:

    * ``'average'`` : The averaged score is computed using the :py:func:`numpy.average` function.
    * ``'min'`` : The minimum score is computed using the :py:func:`min` function.
    * ``'max'`` : The maximum score is computed using the :py:func:`max` function.
    * ``'median'`` : The median score is computed using the :py:func:`numpy.median` function.
    * ``None`` is also accepted, in which case ``None`` is returned.
    """
    try:
        return {
            "average": numpy.average,
            "min": min,
            "max": max,
            "median": numpy.median,
            None: None,
        }[strategy_name]
    except KeyError:
        #    warn("score fusion strategy '%s' is unknown" % strategy_name)
        return None


def selected_indices(total_number_of_indices, desired_number_of_indices=None):
    """Returns a list of indices that will contain exactly the number of desired indices (or the number of total items in the list, if this is smaller).
    These indices are selected such that they are evenly spread over the whole sequence.
    """
    if (
        desired_number_of_indices is None
        or desired_number_of_indices >= total_number_of_indices
        or desired_number_of_indices < 0
    ):
        return range(total_number_of_indices)
    increase = float(total_number_of_indices) / float(desired_number_of_indices)
    # generate a regular quasi-random index list
    return [int((i + 0.5) * increase) for i in range(desired_number_of_indices)]


def selected_elements(list_of_elements, desired_number_of_elements=None):
    """Returns a list of elements that are sub-selected from the given list (or the list itself, if its length is smaller).
    These elements are selected such that they are evenly spread over the whole list.
    """
    total_number_of_elements = len(list_of_elements)
    if (
        desired_number_of_elements is None
        or desired_number_of_elements >= total_number_of_elements
        or desired_number_of_elements < 0
    ):
        return list_of_elements
    # sub-select
    return [
        list_of_elements[i]
        for i in selected_indices(
            total_number_of_elements, desired_number_of_elements
        )
    ]


def pretty_print(obj, kwargs):
    """Returns a pretty-print of the parameters to the constructor of a class, which should be able to copy-paste on the command line to create the object (with few exceptions)."""
    return "%s(%s)" % (
        str(obj.__class__),
        ", ".join(
            [
                "%s='%s'" % (key, value)
                if isinstance(value, str)
                else "%s=%s" % (key, value)
                for key, value in kwargs.items()
                if value is not None
            ]
        ),
    )


def is_argument_available(argument, method):
    """
    Check if an argument (or keyword argument) is available in a method

    Attributes
    ----------
      argument: str
        The name of the argument (or keyword argument).

      method:
        Pointer to the method

    """
    return argument in inspect.signature(method).parameters.keys()



# isort: skip_file
from .load import (  # noqa: F401
    open_file,
    split,
    cmc,
    split_four_column,
    four_column,
    split_five_column,
    five_column,
    scores,
    load_score,
    dump_score,
    cmc_four_column,
    cmc_five_column,
    split_csv_vuln,
    iscsv,
)



import os

import bob.io.base

from bob.bio.base.config.dummy.database import database

# Creates a list of unique str
samples = [s.key for s in database.background_model_samples()]


def reader(sample):
    data = bob.io.base.load(
        os.path.join(
            database.original_directory, sample + database.original_extension
        )
    )
    return data


def make_key(sample):
    return sample



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Manuel Guenther <Manuel.Guenther@idiap.ch>
# @date: Tue Oct  2 12:12:39 CEST 2012

import warnings

from .. import utils


class Extractor(object):
    """This is the base class for all feature extractors.
    It defines the minimum requirements that a derived feature extractor class need to implement.

    If your derived class requires training, please register this here.

    **Parameters**

    requires_training : bool
      Set this flag to ``True`` if your feature extractor needs to be trained.
      In that case, please override the :py:meth:`train` and :py:meth:`load` methods

    split_training_data_by_client : bool
      Set this flag to ``True`` if your feature extractor requires the training data to be split by clients.
      Ignored, if ``requires_training`` is ``False``

    min_extractor_file_size : int
        The minimum file size of a saved extractor file for extractors that
        require training in bytes. If the saved file size is smaller than this, it
        is assumed to be a corrupt file and the extractor will be trained again.

    min_feature_file_size : int
        The minimum file size of extracted features in bytes. If the saved file
        size is smaller than this, it is assumed to be a corrupt file and the
        features will be extracted again.

    kwargs : ``key=value`` pairs
      A list of keyword arguments to be written in the `__str__` function.
    """

    def __init__(
        self,
        requires_training=False,  # enable, if your extractor needs training
        split_training_data_by_client=False,  # enable, if your extractor needs the training files sorted by client
        min_extractor_file_size=1000,
        min_feature_file_size=1000,
        **kwargs,  # the parameters of the extractor, to be written in the __str__() method
    ):
        # Each class needs to have a constructor taking
        # all the parameters that are required for the feature extraction as arguments
        self.requires_training = requires_training
        self.split_training_data_by_client = split_training_data_by_client
        self.min_extractor_file_size = min_extractor_file_size
        self.min_feature_file_size = min_feature_file_size
        self._kwargs = kwargs

        warnings.warn(
            "`bob.bio.base.extractor.Extractor` will be deprecated in 01/01/2021. "
            "Please, implement your biometric algorithm using `bob.pipelines` (https://gitlab.idiap.ch/bob/bob.pipelines).",
            DeprecationWarning,
        )

    ############################################################
    # functions that must be overwritten in derived classes
    ############################################################

    def __call__(self, data):
        """__call__(data) -> feature

        This function will actually perform the feature extraction.
        It must be overwritten by derived classes.
        It takes the (preprocessed) data and returns the features extracted from the data.

        **Parameters**

        data : object (usually :py:class:`numpy.ndarray`)
          The *preprocessed* data from which features should be extracted.

        **Returns:**

        feature : object (usually :py:class:`numpy.ndarray`)
          The extracted feature.
        """
        raise NotImplementedError(
            "Please overwrite this function in your derived class"
        )

    def __str__(self):
        """__str__() -> info

        This function returns all parameters of this class (and its derived class).

        **Returns:**

        info : str
          A string containing the full information of all parameters of this (and the derived) class.
        """
        return utils.pretty_print(self, self._kwargs)

    ############################################################
    # Special functions that might be overwritten on need
    ############################################################

    def write_feature(self, feature, feature_file):
        """Writes the given *extracted* feature to a file with the given name.
        In this base class implementation, we simply use :py:func:`bob.bio.base.save` for that.
        If you have a different format, please overwrite this function.

        **Parameters:**

        feature : object
          The extracted feature, i.e., what is returned from `__call__`.

        feature_file : str or :py:class:`h5py.File`
          The file open for writing, or the name of the file to write.
        """
        utils.save(feature, feature_file)

    def read_feature(self, feature_file):
        """Reads the *extracted* feature from file.
        In this base class implementation, it uses :py:func:`bob.bio.base.load` to do that.
        If you have different format, please overwrite this function.

        **Parameters:**

        feature_file : str or :py:class:`h5py.File`
          The file open for reading or the name of the file to read from.

        **Returns:**

        feature : object (usually :py:class:`numpy.ndarray`)
          The feature read from file.
        """
        return utils.load(feature_file)

    def load(self, extractor_file):
        """Loads the parameters required for feature extraction from the extractor file.
        This function usually is only useful in combination with the :py:meth:`train` function.
        In this base class implementation, it does nothing.

        **Parameters:**

        extractor_file : str
          The file to read the extractor from.
        """
        pass

    def train(self, training_data, extractor_file):
        """This function can be overwritten to train the feature extractor.
        If you do this, please also register the function by calling this base class constructor
        and enabling the training by ``requires_training = True``.

        **Parameters:**

        training_data : [object] or [[object]]
          A list of *preprocessed* data that can be used for training the extractor.
          Data will be provided in a single list, if ``split_training_features_by_client = False`` was specified in the constructor,
          otherwise the data will be split into lists, each of which contains the data of a single (training-)client.

        extractor_file : str
          The file to write.
          This file should be readable with the :py:meth:`load` function.
        """
        raise NotImplementedError(
            "Please overwrite this function in your derived class, or unset the 'requires_training' option in the constructor."
        )



import numpy

from bob.bio.base.database import BioFile
from bob.bio.base.preprocessor import Preprocessor

numpy.random.seed(10)


class DummyPreprocessor(Preprocessor):
    def __init__(self, return_none=False, probability_of_none=1, **kwargs):
        Preprocessor.__init__(self)
        self.return_none = return_none
        self.probability_of_none = probability_of_none

    def __call__(self, data, annotation):
        """Does nothing, simply converts the data type of the data, ignoring any annotation."""
        if self.return_none:
            return numpy.random.choice(
                [None, data],
                p=[self.probability_of_none, 1 - self.probability_of_none],
            )

        return data


preprocessor = DummyPreprocessor()


class DummyPreprocessorMetadata(DummyPreprocessor):
    def __call__(self, data, annotation, metadata=None):
        """Does nothing, simply converts the data type of the data, ignoring any annotation."""
        assert isinstance(metadata, BioFile)
        return super(DummyPreprocessorMetadata, self).__call__(data, annotation)


preprocessor_metadata = DummyPreprocessorMetadata()



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
import logging
import os

from bob.bio.base.database.legacy import check_parameters_for_validity
from bob.bio.base.utils.annotations import read_annotation_file

from .. import BioFile, ZTBioDatabase
from .models import ListReader

logger = logging.getLogger("bob.bio.base")


class FileListBioDatabase(ZTBioDatabase):
    """This class provides a user-friendly interface to databases that are given as file lists.

    Parameters
    ----------

    filelists_directory : str
      The directory that contains the filelists defining the protocol(s). If you use the protocol
      attribute when querying the database, it will be appended to the base directory, such that
      several protocols are supported by the same class instance of `bob.bio.base`.

    name : str
      The name of the database

    protocol : str
      The protocol of the database. This should be a folder inside ``filelists_directory``.

    bio_file_class : ``class``
      The class that should be used to return the files.
      This can be :py:class:`bob.bio.base.database.BioFile`, :py:class:`bob.bio.spear.database.AudioBioFile`, :py:class:`bob.bio.face.database.FaceBioFile`, or anything similar.

    original_directory : str or ``None``
      The directory, where the original data can be found.

    original_extension : str or [str] or ``None``
      The filename extension of the original data, or multiple extensions.

    annotation_directory : str or ``None``
      The directory, where additional annotation files can be found.

    annotation_extension : str or ``None``
      The filename extension of the annotation files.

    annotation_type : str or ``None``
      The type of annotation that can be read.
      Currently, options are ``'eyecenter', 'named', 'idiap'``.
      See :py:func:`read_annotation_file` for details.

    dev_sub_directory : str or ``None``
      Specify a custom subdirectory for the filelists of the development set (default is ``'dev'``)

    eval_sub_directory : str or ``None``
      Specify a custom subdirectory for the filelists of the development set (default is ``'eval'``)

    world_filename : str or ``None``
      Specify a custom filename for the training filelist (default is ``'norm/train_world.lst'``)

    optional_world_1_filename : str or ``None``
      Specify a custom filename for the (first optional) training filelist
      (default is ``'norm/train_optional_world_1.lst'``)

    optional_world_2_filename : str or ``None``
      Specify a custom filename for the (second optional) training filelist
      (default is ``'norm/train_optional_world_2.lst'``)

    models_filename : str or ``None``
      Specify a custom filename for the model filelists (default is ``'for_models.lst'``)

    probes_filename : str or ``None``
      Specify a custom filename for the probes filelists (default is ``'for_probes.lst'``)

    scores_filename : str or ``None``
      Specify a custom filename for the scores filelists (default is ``'for_scores.lst'``)

    tnorm_filename : str or ``None``
      Specify a custom filename for the T-norm scores filelists (default is ``'for_tnorm.lst'``)

    znorm_filename : str or ``None``
      Specify a custom filename for the Z-norm scores filelists (default is ``'for_znorm.lst'``)

    use_dense_probe_file_list : bool or None
      Specify which list to use among ``probes_filename`` (dense) or ``scores_filename``.
      If ``None`` it is tried to be estimated based on the given parameters.

    keep_read_lists_in_memory : bool
      If set to ``True`` (the default), the lists are read only once and stored in memory.
      Otherwise the lists will be re-read for every query (not recommended).
    """

    def __init__(
        self,
        filelists_directory,
        name,
        protocol=None,
        bio_file_class=BioFile,
        original_directory=None,
        original_extension=None,
        annotation_directory=None,
        annotation_extension=".json",
        annotation_type="json",
        dev_sub_directory=None,
        eval_sub_directory=None,
        world_filename=None,
        optional_world_1_filename=None,
        optional_world_2_filename=None,
        models_filename=None,
        # For probing, use ONE of the two score file lists:
        probes_filename=None,  # File containing the probe files -> dense model/probe score matrix
        scores_filename=None,  # File containing list of model and probe files -> sparse model/probe score matrix
        # For ZT-Norm:
        tnorm_filename=None,
        znorm_filename=None,
        use_dense_probe_file_list=None,
        # if both probe_filename and scores_filename is given, what kind of list should be used?
        keep_read_lists_in_memory=True,
        # if set to True (the RECOMMENDED default) lists are read only once and stored in memory.
        **kwargs,
    ):
        """Initializes the database with the file lists from the given base directory,
        and the given sub-directories and file names (which default to useful values if not given).
        """

        super(FileListBioDatabase, self).__init__(
            name=name,
            protocol=protocol,
            original_directory=original_directory,
            original_extension=original_extension,
            annotation_directory=annotation_directory,
            annotation_extension=annotation_extension,
            annotation_type=annotation_type,
            **kwargs,
        )
        # extra args for pretty printing
        self._kwargs.update(
            dict(
                filelists_directory=filelists_directory,
                dev_sub_directory=dev_sub_directory,
                eval_sub_directory=eval_sub_directory,
                world_filename=world_filename,
                optional_world_1_filename=optional_world_1_filename,
                optional_world_2_filename=optional_world_2_filename,
                models_filename=models_filename,
                probes_filename=probes_filename,
                scores_filename=scores_filename,
                tnorm_filename=tnorm_filename,
                znorm_filename=znorm_filename,
                use_dense_probe_file_list=use_dense_probe_file_list,
                # if both probe_filename and scores_filename are given, what kind
                # of list should be used?
                keep_read_lists_in_memory=keep_read_lists_in_memory,
            )
        )
        # self.original_directory = original_directory
        # self.original_extension = original_extension
        self.bio_file_class = bio_file_class
        self.keep_read_lists_in_memory = keep_read_lists_in_memory
        self.list_readers = {}

        self.m_base_dir = os.path.abspath(filelists_directory)
        if not os.path.isdir(self.m_base_dir):
            raise RuntimeError(
                "Invalid directory specified %s." % (self.m_base_dir)
            )

        # sub-directories for dev and eval set:
        self.m_dev_subdir = (
            dev_sub_directory if dev_sub_directory is not None else "dev"
        )
        self.m_eval_subdir = (
            eval_sub_directory if eval_sub_directory is not None else "eval"
        )

        # training list:     format:   filename client_id
        self.m_world_filename = (
            world_filename
            if world_filename is not None
            else os.path.join("norm", "train_world.lst")
        )
        # optional training list 1:     format:   filename client_id
        self.m_optional_world_1_filename = (
            optional_world_1_filename
            if optional_world_1_filename is not None
            else os.path.join("norm", "train_optional_world_1.lst")
        )
        # optional training list 2:     format:   filename client_id
        self.m_optional_world_2_filename = (
            optional_world_2_filename
            if optional_world_2_filename is not None
            else os.path.join("norm", "train_optional_world_2.lst")
        )
        # model list:        format:   filename model_id client_id
        self.m_models_filename = (
            models_filename if models_filename is not None else "for_models.lst"
        )
        # scores list:       format:   filename model_id claimed_client_id client_id
        self.m_scores_filename = (
            scores_filename if scores_filename is not None else "for_scores.lst"
        )
        # probe list:        format:   filename client_id
        self.m_probes_filename = (
            probes_filename if probes_filename is not None else "for_probes.lst"
        )
        # T-Norm models      format:   filename model_id client_id
        self.m_tnorm_filename = (
            tnorm_filename if tnorm_filename is not None else "for_tnorm.lst"
        )
        # Z-Norm files       format:   filename client_id
        self.m_znorm_filename = (
            znorm_filename if znorm_filename is not None else "for_znorm.lst"
        )

        self.m_use_dense_probe_file_list = use_dense_probe_file_list

    def _list_reader(self, protocol):
        if protocol not in self.list_readers:
            if protocol is not None:
                protocol_dir = os.path.join(self.get_base_directory(), protocol)
                if not os.path.isdir(protocol_dir):
                    raise ValueError(
                        "The directory %s for the given protocol '%s' does not exist"
                        % (protocol_dir, protocol)
                    )
            self.list_readers[protocol] = ListReader(
                self.keep_read_lists_in_memory
            )

        return self.list_readers[protocol]

    def _make_bio(self, files):
        return [
            self.bio_file_class(
                client_id=f.client_id, path=f.path, file_id=f.id
            )
            for f in files
        ]

    def all_files(self, groups=["dev"], add_zt_files=True):
        """Returns all files for the given group. The internally stored protocol is used, throughout.

        Parameters
        ----------

        groups : [str]
          A list of groups to retrieve the files for.

        add_zt_files : bool
          If selected, also files for ZT-norm scoring will be added.
          Please select this option only if this dataset provides ZT-norm files, see :py:meth:`implements_zt`.

        Returns
        -------

        [BioFile]
          A list of all files that fulfill your query.
        """
        files = self.objects(groups, self.protocol, **self.all_files_options)
        # add all files that belong to the ZT-norm
        for group in groups:
            if group == "world":
                continue
            if add_zt_files:
                if self.implements_zt(self.protocol, group):
                    files += self.tobjects(group, self.protocol)
                    files += self.zobjects(
                        group, self.protocol, **self.z_probe_options
                    )
                else:
                    logger.warning(
                        "ZT score files are requested, but no such files are defined in group %s for protocol %s",
                        group,
                        self.protocol,
                    )

        return self.sort(self._make_bio(files))

    def groups(self, protocol=None, add_world=True, add_subworld=True):
        """This function returns the list of groups for this database.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol for which the groups should be retrieved.
          If ``None``, the internally stored protocol is used.

        add_world : bool
          Add the world groups?

        add_subworld : bool
          Add the sub-world groups? Only valid, when ``add_world=True``

        Returns
        -------

        [str]
          A list of groups
        """
        groups = []
        protocol = protocol or self.protocol
        if protocol is not None:
            if os.path.isdir(
                os.path.join(
                    self.get_base_directory(), protocol, self.m_dev_subdir
                )
            ):
                groups.append("dev")
            if os.path.isdir(
                os.path.join(
                    self.get_base_directory(), protocol, self.m_eval_subdir
                )
            ):
                groups.append("eval")
            if add_world:
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(),
                        protocol,
                        self.m_world_filename,
                    )
                ):
                    groups.append("world")
            if add_world and add_subworld:
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(),
                        protocol,
                        self.m_optional_world_1_filename,
                    )
                ):
                    groups.append("optional_world_1")
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(),
                        protocol,
                        self.m_optional_world_2_filename,
                    )
                ):
                    groups.append("optional_world_2")
        else:
            if os.path.isdir(
                os.path.join(self.get_base_directory(), self.m_dev_subdir)
            ):
                groups.append("dev")
            if os.path.isdir(
                os.path.join(self.get_base_directory(), self.m_eval_subdir)
            ):
                groups.append("eval")
            if add_world:
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(), self.m_world_filename
                    )
                ):
                    groups.append("world")
            if add_world and add_subworld:
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(),
                        self.m_optional_world_1_filename,
                    )
                ):
                    groups.append("optional_world_1")
                if os.path.isfile(
                    os.path.join(
                        self.get_base_directory(),
                        self.m_optional_world_2_filename,
                    )
                ):
                    groups.append("optional_world_2")
        return groups

    def implements_zt(self, protocol=None, groups=None):
        """Checks if the file lists for the ZT score normalization are available.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol for which the groups should be retrieved.

        groups : str or [str] or ``None``
          The groups for which the ZT score normalization file lists should be checked ``('dev', 'eval')``.

        Returns
        -------

        bool
          ``True`` if the all file lists for ZT score normalization exist, otherwise ``False``.
        """
        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        for group in groups:
            for t in ["for_tnorm", "for_znorm"]:
                if not os.path.exists(self._get_list_file(group, t, protocol)):
                    return False
        # all files exist
        return True

    def uses_dense_probe_file(self, protocol):
        """Determines if a dense probe file list is used based on the existence of parameters."""
        # return, whatever was specified in constructor, if not None
        if self.m_use_dense_probe_file_list is not None:
            return self.m_use_dense_probe_file_list

        # check the existence of the files
        probes = True
        scores = True
        for group in self.groups(protocol, add_world=False):
            probes = probes and os.path.exists(
                self._get_list_file(group, type="for_probes", protocol=protocol)
            )
            scores = scores and os.path.exists(
                self._get_list_file(group, type="for_scores", protocol=protocol)
            )
        # decide, which score files are available
        if probes and not scores:
            return True
        if not probes and scores:
            return False
        raise ValueError(
            "Unable to determine, which way of probing should be used. Please specify."
        )

    def get_base_directory(self):
        """Returns the base directory where the filelists defining the database
        are located."""
        return self.m_base_dir

    def set_base_directory(self, filelists_directory):
        """Resets the base directory where the filelists defining the database
        are located."""
        self.m_base_dir = filelists_directory
        if not os.path.isdir(self.filelists_directory):
            raise RuntimeError(
                "Invalid directory specified %s." % (self.filelists_directory)
            )

    def _get_list_file(self, group, type=None, protocol=None):
        if protocol:
            base_directory = os.path.join(self.get_base_directory(), protocol)
        else:
            base_directory = self.get_base_directory()
        if group == "world":
            return os.path.join(base_directory, self.m_world_filename)
        elif group == "optional_world_1":
            return os.path.join(
                base_directory, self.m_optional_world_1_filename
            )
        elif group == "optional_world_2":
            return os.path.join(
                base_directory, self.m_optional_world_2_filename
            )
        else:
            group_dir = (
                self.m_dev_subdir if group == "dev" else self.m_eval_subdir
            )
            list_name = {
                "for_models": self.m_models_filename,
                "for_probes": self.m_probes_filename,
                "for_scores": self.m_scores_filename,
                "for_tnorm": self.m_tnorm_filename,
                "for_znorm": self.m_znorm_filename,
            }[type]
            return os.path.join(base_directory, group_dir, list_name)

    def client_id_from_model_id(self, model_id, group="dev"):
        """Returns the client id that is connected to the given model id.

        Parameters
        ----------

        model_id : str or ``None``
          The model id for which the client id should be returned.

        groups : str or [str] or ``None``
          (optional) the groups, the client belongs to.
          Might be one or more of ``('dev', 'eval', 'world', 'optional_world_1', 'optional_world_2')``.
          If groups are given, only these groups are considered.

        protocol : str or ``None``
          The protocol to consider.

        Returns
        -------

        str
          The client id for the given model id, if found.
        """
        protocol = self.protocol
        groups = check_parameters_for_validity(
            group,
            "group",
            self.groups(protocol),
            default_parameters=self.groups(protocol, add_subworld=False),
        )

        for group in groups:
            model_dict = self._list_reader(protocol).read_models(
                self._get_list_file(group, "for_models", protocol),
                group,
                "for_models",
            )
            if model_id in model_dict:
                return model_dict[model_id]

        raise ValueError(
            "The given model id '%s' cannot be found in one of the groups '%s'"
            % (model_id, groups)
        )

    def client_id_from_t_model_id(self, t_model_id, group="dev"):
        """Returns the client id that is connected to the given T-Norm model id.

        Parameters
        ----------

        model_id : str or ``None``
          The model id for which the client id should be returned.

        groups : str or [str] or ``None``
          (optional) the groups, the client belongs to.
          Might be one or more of ``('dev', 'eval')``.
          If groups are given, only these groups are considered.

        Returns
        -------

        str
          The client id for the given model id of a T-Norm model, if found.
        """
        protocol = self.protocol
        groups = check_parameters_for_validity(
            group, "group", self.groups(protocol, add_world=False)
        )

        for group in groups:
            model_dict = self._list_reader(protocol).read_models(
                self._get_list_file(group, "for_tnorm", protocol),
                group,
                "for_tnorm",
            )
            if t_model_id in model_dict:
                return model_dict[t_model_id]

        raise ValueError(
            "The given T-norm model id '%s' cannot be found in one of the groups '%s'"
            % (t_model_id, groups)
        )

    def __client_id_list__(self, groups, type, protocol=None):
        ids = set()
        protocol = protocol or self.protocol
        # read all lists for all groups and extract the model ids
        for group in groups:
            files = self._list_reader(protocol).read_list(
                self._get_list_file(group, type, protocol), group, type
            )
            for file in files:
                ids.add(file.client_id)
        return ids

    def client_ids(self, protocol=None, groups=None):
        """Returns a list of client ids for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the clients belong ``('dev', 'eval', 'world', 'optional_world_1', 'optional_world_2')``.

        Returns
        -------

        [str]
          A list containing all the client ids which have the given properties.
        """

        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups,
            "group",
            self.groups(protocol),
            default_parameters=self.groups(protocol, add_subworld=False),
        )

        return self.__client_id_list__(groups, "for_models", protocol)

    def tclient_ids(self, protocol=None, groups=None):
        """Returns a list of T-Norm client ids for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the clients belong ("dev", "eval").

        Returns
        -------

        [str]
          A list containing all the T-Norm client ids which have the given properties.
        """

        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        return self.__client_id_list__(groups, "for_tnorm", protocol)

    def zclient_ids(self, protocol=None, groups=None):
        """Returns a list of Z-Norm client ids for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the clients belong ("dev", "eval").

        Returns
        -------

        [str]
          A list containing all the Z-Norm client ids which have the given properties.
        """

        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        return self.__client_id_list__(groups, "for_znorm", protocol)

    def __model_id_list__(self, groups, type, protocol=None):
        ids = set()
        protocol = protocol or self.protocol
        # read all lists for all groups and extract the model ids
        for group in groups:
            dict = self._list_reader(protocol).read_models(
                self._get_list_file(group, type, protocol), group, type
            )
            ids.update(dict.keys())
        return list(ids)

    def model_ids_with_protocol(self, groups=None, protocol=None, **kwargs):
        """Returns a list of model ids for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the models belong ``('dev', 'eval', 'world', 'optional_world_1', 'optional_world_2')``.

        Returns
        -------

        [str]
          A list containing all the model ids which have the given properties.
        """
        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol=protocol)
        )

        return self.__model_id_list__(groups, "for_models", protocol)

    def tmodel_ids_with_protocol(self, protocol=None, groups=None, **kwargs):
        """Returns a list of T-Norm model ids for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the models belong ``('dev', 'eval')``.

        Returns
        -------

        [str]
          A list containing all the T-Norm model ids belonging to the given group.
        """
        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        return self.__model_id_list__(groups, "for_tnorm", protocol)

    def objects(
        self,
        groups=None,
        protocol=None,
        purposes=None,
        model_ids=None,
        classes=None,
        **kwargs,
    ):
        """Returns a set of :py:class:`bob.bio.base.database.BioFile` objects for the specific query by the user.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        purposes : str or [str] or ``None``
          The purposes required to be retrieved ``('enroll', 'probe')`` or a tuple
          with several of them. If ``None`` is given (this is the default), it is
          considered the same as a tuple with all possible values. This field is
          ignored for the data from the ``'world', 'optional_world_1', 'optional_world_2'`` groups.

        model_ids : str or [str] or ``None``
          Only retrieves the files for the provided list of model ids (claimed
          client id). If ``None`` is given (this is the default), no filter over
          the model_ids is performed.

        groups : str or [str] or ``None``
          One of the groups ``('dev', 'eval', 'world', 'optional_world_1', 'optional_world_2')`` or a tuple with several of them.
          If ``None`` is given (this is the default), it is considered to be the existing subset of ``('world', 'dev', 'eval')``.

        classes : str or [str] or ``None``
          The classes (types of accesses) to be retrieved ``('client', 'impostor')``
          or a tuple with several of them. If ``None`` is given (this is the
          default), it is considered the same as a tuple with all possible values.

          .. note::
             Classes are not allowed to be specified when 'probes_filename' is used in the constructor.

        Returns
        -------

        [BioFile]
          A list of :py:class:`BioFile` objects considering all the filtering criteria.
        """

        protocol = protocol or self.protocol
        if self.uses_dense_probe_file(protocol) and classes is not None:
            raise ValueError(
                "To be able to use the 'classes' keyword, please use the 'for_scores.lst' list file."
            )

        purposes = check_parameters_for_validity(
            purposes, "purpose", ("enroll", "probe")
        )
        groups = check_parameters_for_validity(
            groups,
            "group",
            self.groups(protocol),
            default_parameters=self.groups(protocol, add_subworld=False),
        )
        classes = check_parameters_for_validity(
            classes, "class", ("client", "impostor")
        )

        if isinstance(model_ids, str):
            model_ids = (model_ids,)

        # first, collect all the lists that we want to process
        lists = []
        probe_lists = []
        if "world" in groups:
            lists.append(
                self._list_reader(protocol).read_list(
                    self._get_list_file("world", protocol=protocol), "world"
                )
            )
        if "optional_world_1" in groups:
            lists.append(
                self._list_reader(protocol).read_list(
                    self._get_list_file("optional_world_1", protocol=protocol),
                    "optional_world_1",
                )
            )
        if "optional_world_2" in groups:
            lists.append(
                self._list_reader(protocol).read_list(
                    self._get_list_file("optional_world_2", protocol=protocol),
                    "optional_world_2",
                )
            )

        for group in ("dev", "eval"):
            if group in groups:
                if "enroll" in purposes:
                    lists.append(
                        self._list_reader(protocol).read_list(
                            self._get_list_file(
                                group, "for_models", protocol=protocol
                            ),
                            group,
                            "for_models",
                        )
                    )
                if "probe" in purposes:
                    if self.uses_dense_probe_file(protocol):
                        probe_lists.append(
                            self._list_reader(protocol).read_list(
                                self._get_list_file(
                                    group, "for_probes", protocol=protocol
                                ),
                                group,
                                "for_probes",
                            )
                        )
                    else:
                        probe_lists.append(
                            self._list_reader(protocol).read_list(
                                self._get_list_file(
                                    group, "for_scores", protocol=protocol
                                ),
                                group,
                                "for_scores",
                            )
                        )

        # now, go through the lists and filter the elements

        # remember the file ids that are already in the list
        file_ids = set()
        retval = []

        # non-probe files; just filter by model id
        for list in lists:
            for file in list:
                # check if we already have this file
                if file.id not in file_ids:
                    if model_ids is None or file._model_id in model_ids:
                        file_ids.add(file.id)
                        retval.append(file)

        # probe files; filter by model id and by class
        for list in probe_lists:
            if self.uses_dense_probe_file(protocol):
                # dense probing is used; do not filter over the model ids and not over the classes
                # -> just add all probe files
                for file in list:
                    if file.id not in file_ids:
                        file_ids.add(file.id)
                        retval.append(file)

            else:
                # sparse probing is used; filter over model ids and over the classes
                for file in list:
                    # filter by model id
                    if model_ids is None or file._model_id in model_ids:
                        # filter by class
                        if (
                            "client" in classes
                            and file.client_id == file.claimed_id
                        ) or (
                            "impostor" in classes
                            and file.client_id != file.claimed_id
                        ):
                            # check if we already have this file
                            if file.id not in file_ids:
                                file_ids.add(file.id)
                                retval.append(file)

        return self._make_bio(retval)

    def tobjects(self, groups=None, protocol=None, model_ids=None, **kwargs):
        """Returns a list of :py:class:`bob.bio.base.database.BioFile` objects for enrolling T-norm models for score normalization.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        model_ids : str or [str] or ``None``
          Only retrieves the files for the provided list of model ids (claimed
          client id). If ``None`` is given (this is the default), no filter over
          the model_ids is performed.

        groups : str or [str] or ``None``
          The groups to which the models belong ``('dev', 'eval')``.

        Returns
        -------

        [BioFile]
          A list of :py:class:`BioFile` objects considering all the filtering criteria.
        """
        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        if isinstance(model_ids, str):
            model_ids = (model_ids,)

        # iterate over the lists and extract the files
        # we assume that there is no duplicate file here...
        retval = []
        for group in groups:
            for file in self._list_reader(protocol).read_list(
                self._get_list_file(group, "for_tnorm", protocol),
                group,
                "for_tnorm",
            ):
                if model_ids is None or file._model_id in model_ids:
                    retval.append(file)

        return self._make_bio(retval)

    def zobjects(self, groups=None, protocol=None, **kwargs):
        """Returns a list of :py:class:`BioFile` objects to perform Z-norm score normalization.

        Parameters
        ----------

        protocol : str or ``None``
          The protocol to consider

        groups : str or [str] or ``None``
          The groups to which the clients belong ``('dev', 'eval')``.

        Returns
        -------

        [BioFile]
          A list of File objects considering all the filtering criteria.
        """

        protocol = protocol or self.protocol
        groups = check_parameters_for_validity(
            groups, "group", self.groups(protocol, add_world=False)
        )

        # iterate over the lists and extract the files
        # we assume that there is no duplicate file here...
        retval = []
        for group in groups:
            retval.extend(
                [
                    file
                    for file in self._list_reader(protocol).read_list(
                        self._get_list_file(group, "for_znorm", protocol),
                        group,
                        "for_znorm",
                    )
                ]
            )

        return self._make_bio(retval)

    def annotations(self, file):
        """Reads the annotations for the given file id from file and returns them in a dictionary.

        Parameters
        ----------

        file : BioFile
          The BioFile object for which the annotations should be read.

        Returns
        -------

        dict
          The annotations as a dictionary, e.g.: ``{'reye':(re_y,re_x), 'leye':(le_y,le_x)}``
        """
        if self.annotation_directory is None:
            return None

        # since the file id is equal to the file name, we can simply use it
        annotation_file = os.path.join(
            self.annotation_directory, file.id + self.annotation_extension
        )

        # return the annotations as read from file
        return read_annotation_file(annotation_file, self.annotation_type)

    def original_file_name(self, file, check_existence=True):
        """Returns the original file name of the given file.

        This interface supports several original extensions, so that file lists can contain images
        of different data types.

        When multiple original extensions are specified, this function will check the existence of any of
        these file names, and return the first one that actually exists.
        In this case, the ``check_existence`` flag is ignored.

        Parameters
        ----------

        file : BioFile
          The BioFile object for which the file name should be returned.

        check_existence : bool
          Should the existence of the original file be checked?
          (Ignored when multiple original extensions were specified in the constructor.)

        Returns
        -------

        str
          The full path of the original data file.
        """

        if isinstance(self.original_extension, str):
            # extract file name
            file_name = file.make_path(
                self.original_directory, self.original_extension
            )
            if not check_existence or os.path.exists(file_name):
                return file_name

        # check all registered extensions
        for extension in self.original_extension:
            file_name = file.make_path(self.original_directory, extension)
            if os.path.exists(file_name):
                return file_name

        # None of the extensions matched
        raise IOError(
            "File '%s' does not exist with any of the extensions '%s'"
            % (
                file.make_path(self.original_directory, None),
                self.original_extension,
            )
        )



"""Plots and measures for bob.bio.base"""

import logging
import math

import click
import matplotlib.pyplot as mpl

from tabulate import tabulate

import bob.measure
import bob.measure.script.figure as measure_figure

from bob.measure import plot

LOGGER = logging.getLogger("bob.bio.base")


class Roc(measure_figure.Roc):
    def __init__(self, ctx, scores, evaluation, func_load):
        super(Roc, self).__init__(ctx, scores, evaluation, func_load)
        self._x_label = ctx.meta.get("x_label") or "FMR"
        default_y_label = "1 - FNMR" if self._tpr else "FNMR"
        self._y_label = ctx.meta.get("y_label") or default_y_label


class Det(measure_figure.Det):
    def __init__(self, ctx, scores, evaluation, func_load):
        super(Det, self).__init__(ctx, scores, evaluation, func_load)
        self._x_label = ctx.meta.get("x_label") or "FMR (%)"
        self._y_label = ctx.meta.get("y_label") or "FNMR (%)"


class Cmc(measure_figure.PlotBase):
    """Handles the plotting of Cmc"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Cmc, self).__init__(ctx, scores, evaluation, func_load)
        self._semilogx = ctx.meta.get("semilogx", True)
        self._titles = self._titles or ["CMC dev.", "CMC eval."]
        self._x_label = self._x_label or "Rank"
        self._y_label = self._y_label or "Identification rate"
        self._max_R = 0

    def compute(self, idx, input_scores, input_names):
        """Plot CMC for dev and eval data using
        :py:func:`bob.measure.plot.cmc`"""
        mpl.figure(1)
        if self._eval:
            linestyle = "-" if not self._split else self._linestyles[idx]
            LOGGER.info("CMC dev. curve using %s", input_names[0])
            rank = plot.cmc(
                input_scores[0],
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=linestyle,
                label=self._label("dev.", idx),
            )
            self._max_R = max(rank, self._max_R)
            linestyle = "--"
            if self._split:
                mpl.figure(2)
                linestyle = self._linestyles[idx]

            LOGGER.info("CMC eval. curve using %s", input_names[1])
            rank = plot.cmc(
                input_scores[1],
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=linestyle,
                label=self._label("eval.", idx),
            )
            self._max_R = max(rank, self._max_R)
        else:
            LOGGER.info("CMC dev. curve using %s", input_names[0])
            rank = plot.cmc(
                input_scores[0],
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev.", idx),
            )
            self._max_R = max(rank, self._max_R)


class Dir(measure_figure.PlotBase):
    """Handles the plotting of DIR curve"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Dir, self).__init__(ctx, scores, evaluation, func_load)
        self._semilogx = ctx.meta.get("semilogx", True)
        self._rank = ctx.meta.get("rank", 1)
        self._titles = self._titles or ["DIR curve"] * 2
        self._x_label = self._x_label or "False Positive Identification Rate"
        self._y_label = self._y_label or "True Positive Identification Rate"

    def compute(self, idx, input_scores, input_names):
        """Plot DIR for dev and eval data using
        :py:func:`bob.measure.plot.detection_identification_curve`"""
        mpl.figure(1)
        if self._eval:
            linestyle = "-" if not self._split else self._linestyles[idx]
            LOGGER.info("DIR dev. curve using %s", input_names[0])
            plot.detection_identification_curve(
                input_scores[0],
                rank=self._rank,
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=linestyle,
                label=self._label("dev", idx),
            )
            linestyle = "--"
            if self._split:
                mpl.figure(2)
                linestyle = self._linestyles[idx]

            LOGGER.info("DIR eval. curve using %s", input_names[1])
            plot.detection_identification_curve(
                input_scores[1],
                rank=self._rank,
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=linestyle,
                label=self._label("eval", idx),
            )
        else:
            LOGGER.info("DIR dev. curve using %s", input_names[0])
            plot.detection_identification_curve(
                input_scores[0],
                rank=self._rank,
                logx=self._semilogx,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev", idx),
            )

        if self._min_dig is not None:
            mpl.xlim(xmin=math.pow(10, self._min_dig))


class Metrics(measure_figure.Metrics):
    """Compute metrics from score files"""

    def __init__(
        self,
        ctx,
        scores,
        evaluation,
        func_load,
        names=(
            "Failure to Acquire",
            "False Match Rate",
            "False Non Match Rate",
            "False Accept Rate",
            "False Reject Rate",
            "Half Total Error Rate",
        ),
    ):
        super(Metrics, self).__init__(ctx, scores, evaluation, func_load, names)

    def init_process(self):
        if self._criterion == "rr":
            self._thres = (
                [None] * self.n_systems if self._thres is None else self._thres
            )

    def compute(self, idx, input_scores, input_names):
        """Compute metrics for the given criteria"""
        title = self._legends[idx] if self._legends is not None else None
        headers = ["" or title, "Dev. %s" % input_names[0]]
        if self._eval and input_scores[1] is not None:
            headers.append("eval % s" % input_names[1])
        if self._criterion == "rr":
            rr = bob.measure.recognition_rate(input_scores[0], self._thres[idx])
            dev_rr = "%.1f%%" % (100 * rr)
            raws = [["RR", dev_rr]]
            if self._eval and input_scores[1] is not None:
                rr = bob.measure.recognition_rate(
                    input_scores[1], self._thres[idx]
                )
                eval_rr = "%.1f%%" % (100 * rr)
                raws[0].append(eval_rr)
            click.echo(
                tabulate(raws, headers, self._tablefmt), file=self.log_file
            )
        elif self._criterion == "mindcf":
            if "cost" in self._ctx.meta:
                cost = self._ctx.meta.get("cost", 0.99)
            threshold = (
                bob.measure.min_weighted_error_rate_threshold(
                    input_scores[0][0], input_scores[0][1], cost
                )
                if self._thres is None
                else self._thres[idx]
            )
            if self._thres is None:
                click.echo(
                    "[minDCF - Cost:%f] Threshold on Development set `%s`: %e"
                    % (cost, input_names[0], threshold),
                    file=self.log_file,
                )
            else:
                click.echo(
                    "[minDCF] User defined Threshold: %e" % threshold,
                    file=self.log_file,
                )
            # apply threshold to development set
            far, frr = bob.measure.farfrr(
                input_scores[0][0], input_scores[0][1], threshold
            )
            dev_far_str = "%.1f%%" % (100 * far)
            dev_frr_str = "%.1f%%" % (100 * frr)
            dev_mindcf_str = "%.1f%%" % (
                (cost * far + (1 - cost) * frr) * 100.0
            )
            raws = [
                ["FAR", dev_far_str],
                ["FRR", dev_frr_str],
                ["minDCF", dev_mindcf_str],
            ]
            if self._eval and input_scores[1] is not None:
                # apply threshold to development set
                far, frr = bob.measure.farfrr(
                    input_scores[1][0], input_scores[1][1], threshold
                )
                eval_far_str = "%.1f%%" % (100 * far)
                eval_frr_str = "%.1f%%" % (100 * frr)
                eval_mindcf_str = "%.1f%%" % (
                    (cost * far + (1 - cost) * frr) * 100.0
                )
                raws[0].append(eval_far_str)
                raws[1].append(eval_frr_str)
                raws[2].append(eval_mindcf_str)
            click.echo(
                tabulate(raws, headers, self._tablefmt), file=self.log_file
            )
        elif self._criterion == "cllr":
            cllr = bob.measure.calibration.cllr(
                input_scores[0][0], input_scores[0][1]
            )
            min_cllr = bob.measure.calibration.min_cllr(
                input_scores[0][0], input_scores[0][1]
            )
            dev_cllr_str = "%.1f%%" % cllr
            dev_min_cllr_str = "%.1f%%" % min_cllr
            raws = [["Cllr", dev_cllr_str], ["minCllr", dev_min_cllr_str]]
            if self._eval and input_scores[1] is not None:
                cllr = bob.measure.calibration.cllr(
                    input_scores[1][0], input_scores[1][1]
                )
                min_cllr = bob.measure.calibration.min_cllr(
                    input_scores[1][0], input_scores[1][1]
                )
                eval_cllr_str = "%.1f%%" % cllr
                eval_min_cllr_str = "%.1f%%" % min_cllr
                raws[0].append(eval_cllr_str)
                raws[1].append(eval_min_cllr_str)
                click.echo(
                    tabulate(raws, headers, self._tablefmt), file=self.log_file
                )
        else:
            title = self._legends[idx] if self._legends is not None else None
            all_metrics = self._get_all_metrics(idx, input_scores, input_names)
            headers = [" " or title, "Development"]
            rows = [
                [self.names[0], all_metrics[0][0]],
                [self.names[1], all_metrics[0][1]],
                [self.names[2], all_metrics[0][2]],
                [self.names[3], all_metrics[0][3]],
                [self.names[4], all_metrics[0][4]],
                [self.names[5], all_metrics[0][5]],
            ]

            if self._eval:
                # computes statistics for the eval set based on the threshold a
                # priori
                headers.append("Evaluation")
                rows[0].append(all_metrics[1][0])
                rows[1].append(all_metrics[1][1])
                rows[2].append(all_metrics[1][2])
                rows[3].append(all_metrics[1][3])
                rows[4].append(all_metrics[1][4])
                rows[5].append(all_metrics[1][5])

            click.echo(
                tabulate(rows, headers, self._tablefmt), file=self.log_file
            )


class MultiMetrics(measure_figure.MultiMetrics):
    """Compute metrics from score files"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(MultiMetrics, self).__init__(
            ctx,
            scores,
            evaluation,
            func_load,
            names=(
                "Failure to Acquire",
                "False Match Rate",
                "False Non Match Rate",
                "False Accept Rate",
                "False Reject Rate",
                "Half Total Error Rate",
            ),
        )


class Hist(measure_figure.Hist):
    """Histograms for biometric scores"""

    def _setup_hist(self, neg, pos):
        self._title_base = "Biometric scores"
        self._density_hist(pos[0], n=0, label="Genuines", alpha=0.9, color="C2")
        self._density_hist(
            neg[0], n=1, label="Zero-effort impostors", alpha=0.8, color="C0"
        )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
Implementation of the PipelineSimple using Dask :ref:`bob.bio.base.struct_bio_rec_sys`_

This file contains simple processing blocks meant to be used
for bob.bio experiments
"""

import logging

from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline

from bob.bio.base.pipelines.abstract_classes import BioAlgorithm
from bob.pipelines import SampleWrapper, is_instance_nested, wrap

from .score_writers import FourColumnsScoreWriter

logger = logging.getLogger(__name__)
import tempfile


class PipelineSimple:
    """
    The simplest possible pipeline

    This is the backbone of most biometric recognition systems.
    It implements three subpipelines and they are the following:

     - :py:class:`PipelineSimple.train_background_model`: Initializes or trains your transformer.
        It will run :py:meth:`sklearn.base.BaseEstimator.fit`

     - :py:class:`PipelineSimple.enroll_templates`: Creates enrollment templates
        It will run :py:meth:`sklearn.base.BaseEstimator.transform` followed by a sequence of
        :py:meth:`bob.bio.base.pipelines.abstract_classes.BioAlgorithm.create_templates`

     - :py:class:`PipelineSimple.probe_templates`: Creates probe templates
        It will run :py:meth:`sklearn.base.BaseEstimator.transform` followed by a sequence of
        :py:meth:`bob.bio.base.pipelines.abstract_classes.BioAlgorithm.create_templates`

     - :py:class:`PipelineSimple.compute_scores`: Computes scores
        It will run :py:meth:`bob.bio.base.pipelines.abstract_classes.BioAlgorithm.compare`


    Example
    -------
       >>> from sklearn.preprocessing import FunctionTransformer
       >>> from sklearn.pipeline import make_pipeline
       >>> from bob.bio.base.algorithm import Distance
       >>> from bob.bio.base.pipelines import PipelineSimple
       >>> from bob.pipelines import wrap
       >>> import numpy
       >>> linearize = lambda samples: [numpy.reshape(x, (-1,)) for x in samples]
       >>> transformer = wrap(["sample"], FunctionTransformer(linearize))
       >>> transformer_pipeline = make_pipeline(transformer)
       >>> biometric_algorithm = Distance()
       >>> pipeline = PipelineSimple(transformer_pipeline, biometric_algorithm)
       >>> pipeline(samples_for_training_back_ground_model, samplesets_for_enroll, samplesets_for_scoring)  # doctest: +SKIP


    To run this pipeline using Dask, used the function
    :py:func:`dask_bio_pipeline`.

    Example
    -------
      >>> from bob.bio.base.pipelines import dask_bio_pipeline
      >>> pipeline = PipelineSimple(transformer_pipeline, biometric_algorithm)
      >>> pipeline = dask_bio_pipeline(pipeline)
      >>> pipeline(samples_for_training_back_ground_model, samplesets_for_enroll, samplesets_for_scoring).compute()  # doctest: +SKIP


    Parameters
    ----------

    transformer: :py:class`sklearn.pipeline.Pipeline` or a `sklearn.base.BaseEstimator`
        Transformer that will preprocess your data

    biometric_algorithm: :py:class:`bob.bio.base.pipelines.abstract_classes.BioAlgorithm`
        Biometrics algorithm object that implements the methods `enroll` and
        `score` methods

    score_writer: :any:`bob.bio.base.pipelines.ScoreWriter`
        Format to write scores. Default to
        :any:`bob.bio.base.pipelines.FourColumnsScoreWriter`

    """

    def __init__(
        self,
        transformer: Pipeline,
        biometric_algorithm: BioAlgorithm,
        score_writer=None,
    ):
        self.transformer = transformer
        self.biometric_algorithm = biometric_algorithm
        self.score_writer = score_writer
        if self.score_writer is None:
            tempdir = tempfile.TemporaryDirectory()
            self.score_writer = FourColumnsScoreWriter(tempdir.name)

        check_valid_pipeline(self)

    def __call__(
        self,
        background_model_samples,
        biometric_reference_samples,
        probe_samples,
        score_all_vs_all=True,
        return_templates=False,
    ):
        logger.info(" >> PipelineSimple: Training background model")
        self.train_background_model(background_model_samples)

        logger.info(" >> PipelineSimple: Creating enroll templates")
        enroll_templates = self.enroll_templates(biometric_reference_samples)

        logger.info(" >> PipelineSimple: Creating probe templates")
        probe_templates = self.probe_templates(probe_samples)

        logger.info(" >> PipelineSimple: Computing scores")
        scores = self.compute_scores(
            probe_templates,
            enroll_templates,
            score_all_vs_all,
        )

        if return_templates:
            return scores, enroll_templates, probe_templates
        else:
            return scores

    def train_background_model(self, background_model_samples):
        # background_model_samples is a list of Samples

        # We might have algorithms that has no data for training
        if len(background_model_samples) > 0:
            self.transformer.fit(background_model_samples)
        else:
            logger.warning(
                "There's no data to train background model. "
                "For the rest of the execution it will be assumed that the pipeline does not require fit."
            )
        return self.transformer

    def enroll_templates(self, biometric_reference_samples):
        biometric_reference_features = self.transformer.transform(
            biometric_reference_samples
        )

        enroll_templates = (
            self.biometric_algorithm.create_templates_from_samplesets(
                biometric_reference_features, enroll=True
            )
        )

        # a list of Samples
        return enroll_templates

    def probe_templates(self, probe_samples):
        probe_features = self.transformer.transform(probe_samples)

        probe_templates = (
            self.biometric_algorithm.create_templates_from_samplesets(
                probe_features, enroll=False
            )
        )

        # a list of Samples
        return probe_templates

    def compute_scores(
        self,
        probe_templates,
        enroll_templates,
        score_all_vs_all,
    ):
        return self.biometric_algorithm.score_sample_templates(
            probe_templates, enroll_templates, score_all_vs_all
        )

    def write_scores(self, scores):
        if self.score_writer is None:
            raise ValueError("No score writer defined in the pipeline")
        return self.score_writer.write(scores)

    def post_process(self, score_paths, filename):
        if self.score_writer is None:
            raise ValueError("No score writer defined in the pipeline")

        return self.score_writer.post_process(score_paths, filename)


def check_valid_pipeline(pipeline_simple):
    """
    Applying some checks in the PipelineSimple
    """

    # CHECKING THE TRANSFORMER
    # Checking if it's a Scikit Pipeline or an estimator
    if isinstance(pipeline_simple.transformer, Pipeline):
        # Checking if all steps are wrapped as samples, if not, we should wrap them
        for p in pipeline_simple.transformer:
            if (
                not is_instance_nested(p, "estimator", SampleWrapper)
                and type(p) is not str
                and p is not None
            ):
                wrap(["sample"], p)

    # In this case it can be a simple estimator. AND
    # Checking if it's sample wrapper, if not, do it
    elif is_instance_nested(
        pipeline_simple.transformer, "estimator", BaseEstimator
    ) and is_instance_nested(
        pipeline_simple.transformer, "estimator", BaseEstimator
    ):
        wrap(["sample"], pipeline_simple.transformer)
    else:
        raise ValueError(
            f"pipeline_simple.transformer should be instance of either `sklearn.pipeline.Pipeline` or"
            f"sklearn.base.BaseEstimator, not {pipeline_simple.transformer}"
        )

    # Checking the Biometric algorithm
    if not isinstance(pipeline_simple.biometric_algorithm, BioAlgorithm):
        raise ValueError(
            f"pipeline_simple.biometric_algorithm should be instance of `BioAlgorithm`"
            f"not {pipeline_simple.biometric_algorithm}"
        )

    return True



"""
#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

Implementation of a pipeline that post process scores


"""

import logging
import os

from functools import partial
from typing import Union

import dask.dataframe
import numpy as np
import sklearn.pipeline

from scipy.optimize import curve_fit
from scipy.special import expit
from scipy.stats import beta, gamma, weibull_min
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.linear_model import LogisticRegression

import bob.bio.base

from bob.bio.base.score.load import get_split_dataframe
from bob.pipelines import Sample, SampleSet, getattr_nested, is_instance_nested

from .pipelines import PipelineSimple

logger = logging.getLogger(__name__)


class PipelineScoreNorm:
    """
    Apply Z, T or ZT Score normalization on top of Pimple Pipeline

    Reference bibliography from: A Generative Model for Score Normalization in
    Speaker Recognition https://arxiv.org/pdf/1709.09868.pdf


    Example
    -------
       >>> from sklearn.preprocessing import FunctionTransformer
       >>> from sklearn.pipeline import make_pipeline
       >>> from bob.bio.base.algorithm import Distance
       >>> from bob.bio.base.pipelines import PipelineSimple, PipelineScoreNorm, ZNormScores
       >>> from bob.pipelines import wrap
       >>> import numpy
       >>> linearize = lambda samples: [numpy.reshape(x, (-1,)) for x in samples]
       >>> transformer = wrap(["sample"], FunctionTransformer(linearize))
       >>> transformer_pipeline = make_pipeline(transformer)
       >>> biometric_algorithm = Distance()
       >>> pipeline_simple = PipelineSimple(transformer_pipeline, biometric_algorithm)
       >>> z_norm_postprocessor = ZNormScores()
       >>> z_pipeline = PipelineScoreNorm(pipeline_simple, z_norm_postprocessor)
       >>> zt_pipeline(...) #doctest: +SKIP

    Parameters
    ----------
    pipeline_simple
        An instance of :class:`PipelineSimple` to be wrapped with score normalization.

    post_processor
        Transformer that will post process the scores

    score_writer
        A ScoreWriter to write the scores
    """

    def __init__(
        self,
        pipeline_simple: PipelineSimple,
        post_processor: Union[sklearn.pipeline.Pipeline, BaseEstimator],
    ):
        self.pipeline_simple = pipeline_simple
        self.post_processor = post_processor

        # TODO: ACTIVATE THAT
        # check_valid_pipeline(self)

    @property
    def biometric_algorithm(self):
        return self.pipeline_simple.biometric_algorithm

    @biometric_algorithm.setter
    def biometric_algorithm(self, value):
        self.pipeline_simple.biometric_algorithm = value

    @property
    def transformer(self):
        return self.pipeline_simple.transformer

    @transformer.setter
    def transformer(self, value):
        self.pipeline_simple.transformer = value

    @property
    def score_writer(self):
        return self.pipeline_simple.score_writer

    @score_writer.setter
    def score_writer(self, value):
        self.pipeline_simple.score_writer = value

    def __call__(
        self,
        background_model_samples,
        biometric_reference_samples,
        probe_samples,
        post_process_samples,
        score_all_vs_all=False,
    ):
        raw_scores, enroll_templates, probe_templates = self.pipeline_simple(
            background_model_samples=background_model_samples,
            biometric_reference_samples=biometric_reference_samples,
            probe_samples=probe_samples,
            score_all_vs_all=score_all_vs_all,
            return_templates=True,
        )

        # TODO: I know this is ugly, but I don't want to create one pipeline for every single
        # normalization strategy
        if is_instance_nested(
            self,
            "biometric_algorithm",
            bob.bio.base.pipelines.BioAlgCheckpointWrapper,
        ):
            if is_instance_nested(
                self,
                "biometric_algorithm",
                bob.bio.base.pipelines.BioAlgDaskWrapper,
            ):
                self.biometric_algorithm.biometric_algorithm.biometric_reference_dir = os.path.join(
                    self.biometric_algorithm.biometric_algorithm.biometric_reference_dir,
                    "score-norm",
                )

            else:
                self.biometric_algorithm.biometric_reference_dir = os.path.join(
                    self.biometric_algorithm.biometric_reference_dir,
                    "score-norm",
                )

        template_format = getattr_nested(
            self.post_processor, "post_process_template"
        )
        if template_format == "probe":
            post_process_templates = self.pipeline_simple.probe_templates(
                post_process_samples
            )
            post_process_scores = self.pipeline_simple.compute_scores(
                probe_templates=post_process_templates,
                enroll_templates=enroll_templates,
                score_all_vs_all=False,
            )
        elif template_format == "enroll":
            post_process_templates = self.pipeline_simple.enroll_templates(
                post_process_samples
            )
            post_process_scores = self.pipeline_simple.compute_scores(
                probe_templates=probe_templates,
                enroll_templates=post_process_templates,
                score_all_vs_all=True,
            )
        else:
            raise ValueError(
                "post_process_template must be either 'probe' or 'enroll', got {}".format(
                    template_format
                )
            )

        self.post_processor.fit(post_process_scores)
        post_processed_scores = self.post_processor.transform(raw_scores)

        return raw_scores, post_processed_scores

    def write_scores(self, scores):
        return self.pipeline_simple.write_scores(scores)

    def post_process(self, score_paths, filename):
        return self.pipeline_simple.post_process(score_paths, filename)


def copy_learned_attributes(from_estimator, to_estimator):
    attrs = {k: v for k, v in vars(from_estimator).items() if k.endswith("_")}

    for k, v in attrs.items():
        setattr(to_estimator, k, v)


class ZNormScores(TransformerMixin, BaseEstimator):
    """
    Apply Z-Norm Score normalization on top of Simple Pipeline

    Reference bibliography from: A Generative Model for Score Normalization in Speaker Recognition
    https://arxiv.org/pdf/1709.09868.pdf


    Parameters
    ----------

    """

    post_process_template = "probe"

    def __init__(
        self,
        top_norm=False,
        top_norm_score_fraction=0.8,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.top_norm = top_norm
        self.top_norm_score_fraction = top_norm_score_fraction

    def fit(self, z_scores, y=None):
        # TODO: THIS IS SUPER INNEFICIENT, BUT
        # IT'S THE MOST READABLE SOLUTION

        # Stacking scores by biometric reference
        self.z_stats = dict()
        for sset in z_scores:
            for s in sset:
                if s.template_id not in self.z_stats:
                    self.z_stats[s.template_id] = Sample([], parent=s)

                self.z_stats[s.template_id].data.append(s.data)

        # Now computing the statistics in place

        for key in self.z_stats:
            data = self.z_stats[key].data

            # Selecting the top scores
            if self.top_norm:
                # Sorting in ascending order
                data = -np.sort(-data)
                proportion = int(
                    np.floor(len(data) * self.top_norm_score_fraction)
                )
                data = data[0:proportion]

            self.z_stats[key].mu = np.mean(self.z_stats[key].data)
            self.z_stats[key].std = np.std(self.z_stats[key].data)
            # self._z_stats[key].std = legacy_std(
            #    self._z_stats[key].mu, self._z_stats[key].data
            # )
            self.z_stats[key].data = []

        return self

    def transform(self, X):
        if len(X) <= 0:
            # Nothing to be transformed
            return []

        def _transform_samples(X):
            scores = []
            for no_normed_score in X:
                score = (
                    no_normed_score.data
                    - self.z_stats[no_normed_score.template_id].mu
                ) / self.z_stats[no_normed_score.template_id].std

                z_score = Sample(score, parent=no_normed_score)
                scores.append(z_score)
            return scores

        if isinstance(X[0], SampleSet):
            z_normed_scores = []
            # Transforming either Samples or SampleSets
            for probe_scores in X:
                z_normed_scores.append(
                    SampleSet(
                        _transform_samples(probe_scores), parent=probe_scores
                    )
                )
        else:
            # If it is Samples
            z_normed_scores = _transform_samples(X)

        return z_normed_scores


class TNormScores(TransformerMixin, BaseEstimator):
    """
    Apply T-Norm Score normalization on top of Simple Pipeline

    Reference bibliography from: A Generative Model for Score Normalization in Speaker Recognition
    https://arxiv.org/pdf/1709.09868.pdf


    Parameters
    ----------

    """

    post_process_template = "enroll"

    def __init__(
        self,
        top_norm=False,
        top_norm_score_fraction=0.8,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.top_norm = top_norm
        self.top_norm_score_fraction = top_norm_score_fraction

    def fit(self, t_scores, y=None):
        # TODO: THIS IS SUPER INNEFICIENT, BUT
        # IT'S THE MOST READABLE SOLUTION
        # Stacking scores by biometric reference
        self.t_stats = dict()

        for sset in t_scores:
            self.t_stats[sset.template_id] = Sample(
                [s.data for s in sset], parent=sset
            )

        # Now computing the statistics in place
        for key in self.t_stats:
            data = self.t_stats[key].data

            # Selecting the top scores
            if self.top_norm:
                # Sorting in ascending order
                data = -np.sort(-data)
                proportion = int(
                    np.floor(len(data) * self.top_norm_score_fraction)
                )
                data = data[0:proportion]

            self.t_stats[key].mu = np.mean(self.t_stats[key].data)
            self.t_stats[key].std = np.std(self.t_stats[key].data)
            # self._z_stats[key].std = legacy_std(
            #    self._z_stats[key].mu, self._z_stats[key].data
            # )
            self.t_stats[key].data = []

        return self

    def transform(self, X):
        if len(X) <= 0:
            # Nothing to be transformed
            return []

        def _transform_samples(X, stats):
            scores = []
            for no_normed_score in X:
                score = (no_normed_score.data - stats.mu) / stats.std

                t_score = Sample(score, parent=no_normed_score)
                scores.append(t_score)
            return scores

        if isinstance(X[0], SampleSet):
            t_normed_scores = []
            # Transforming either Samples or SampleSets

            for probe_scores in X:
                stats = self.t_stats[probe_scores.template_id]

                t_normed_scores.append(
                    SampleSet(
                        _transform_samples(probe_scores, stats),
                        parent=probe_scores,
                    )
                )
        else:
            # If it is Samples
            t_normed_scores = _transform_samples(X)

        return t_normed_scores


class LLRCalibration(TransformerMixin, BaseEstimator):
    """
    Implements the linear calibration using a logistic function
    defined in:

    `Mandasari, Miranti Indar, et al. "Score calibration in face recognition." Iet Biometrics 3.4 (2014): 246-256.`

    """

    def fit(self, X, y):
        self.regressor = LogisticRegression(
            class_weight="balanced", fit_intercept=True, penalty="l2"
        )
        self.regressor.fit(X, y)
        return self

    def predict_proba(self, X):
        def get_sigmoid_score(X, regressor):
            return expit(X * regressor.coef_ + regressor.intercept_).ravel()

        return get_sigmoid_score(X, self.regressor)


class WeibullCalibration(TransformerMixin, BaseEstimator):
    """
    Implements the weibull calibration using a pair of Weibull
    pdf's defined in:

    `Macarulla Rodriguez, Andrea, Zeno Geradts, and Marcel Worring. "Likelihood Ratios for Deep Neural Networks in Face Comparison." Journal of forensic sciences 65.4 (2020): 1169-1183.`

    """

    def fit(self, X, y):
        def weibull_pdf_genuines(x, c, scale):
            return weibull_min.pdf(np.abs(x), c, scale=scale)

        def weibull_pdf_impostors(x, c, scale):
            return weibull_min.pdf(np.abs(x), c, scale=scale)

        bins = 20

        impostors_X, impostors_Y = np.histogram(
            X[y == 0], bins=bins, density=True
        )
        # averaging the bins
        impostors_Y = 0.5 * (impostors_Y[1:] + impostors_Y[:-1])

        # Binining genuies and impostors for the fit
        genuines_X, genuines_Y = np.histogram(
            X[y == 1], bins=bins, density=True
        )
        # averaging the bins
        genuines_Y = 0.5 * (genuines_Y[1:] + genuines_Y[:-1])

        # Weibull fit for impostor distribution
        impostors_popt, _ = curve_fit(
            weibull_pdf_impostors,
            xdata=impostors_X,
            ydata=impostors_Y,
            p0=[1.0, 1.0],
        )
        impostors_c, impostors_scale = impostors_popt

        # Fitting weibull for genuines and impostors
        genuines_popt, _ = curve_fit(
            weibull_pdf_genuines,
            xdata=genuines_X,
            ydata=genuines_Y,
            p0=[1.0, 1.0],
        )
        genuines_c, genuines_scale = genuines_popt

        self.gen_dist = partial(
            weibull_pdf_genuines, c=genuines_c, scale=genuines_scale
        )
        self.imp_dist = partial(
            weibull_pdf_impostors, c=impostors_c, scale=impostors_scale
        )
        return self

    def predict_proba(self, X):
        epsilon = 1e-10
        return +1 * (
            np.log10(self.gen_dist(X) + epsilon)
            - np.log10(self.imp_dist(X) + epsilon)
        )


class GammaCalibration(TransformerMixin, BaseEstimator):
    """
    Implements the weibull calibration using a pair of  Gamma
    pdf's defined in:

    """

    def fit(self, X, y):
        def gamma_pdf_genuines(x, a, scale):
            return gamma.pdf(np.abs(x), a, scale=scale)

        def gamma_pdf_impostors(x, a, scale):
            return gamma.pdf(np.abs(x), a, scale=scale)

        bins = 100

        impostors_X, impostors_Y = np.histogram(
            X[y == 0], bins=bins, density=True
        )
        # averaging the bins
        impostors_Y = 0.5 * (impostors_Y[1:] + impostors_Y[:-1])

        # Binining genuies and impostors for the fit
        genuines_X, genuines_Y = np.histogram(
            X[y == 1], bins=bins, density=True
        )
        # averaging the bins
        genuines_Y = 0.5 * (genuines_Y[1:] + genuines_Y[:-1])

        # gamma fit for impostor distribution
        impostors_popt, _ = curve_fit(
            gamma_pdf_impostors,
            xdata=impostors_X,
            ydata=impostors_Y,
            p0=[1.0, 1.0],
        )
        impostors_a, impostors_scale = impostors_popt

        # Fitting weibull for genuines and impostors
        genuines_popt, _ = curve_fit(
            gamma_pdf_genuines,
            xdata=genuines_X,
            ydata=genuines_Y,
            p0=[1.0, 1.0],
        )
        genuines_a, genuines_scale = genuines_popt

        self.gen_dist = partial(
            gamma_pdf_genuines, a=genuines_a, scale=genuines_scale
        )
        self.imp_dist = partial(
            gamma_pdf_impostors, a=impostors_a, scale=impostors_scale
        )
        return self

    def predict_proba(self, X):
        epsilon = 1e-10
        return +1 * (
            np.log10(self.gen_dist(X) + epsilon)
            - np.log10(self.imp_dist(X) + epsilon)
        )


class BetaCalibration(TransformerMixin, BaseEstimator):
    """
    Implements the weibull calibration using a pair of  Beta
    pdf's defined in:

    """

    def fit(self, X, y):
        def beta_pdf(x, a, b):
            return beta.pdf(np.abs(x), a, b)

        bins = 100

        impostors_X, impostors_Y = np.histogram(
            X[y == 0], bins=bins, density=True
        )
        # averaging the bins
        impostors_Y = 0.5 * (impostors_Y[1:] + impostors_Y[:-1])

        # Binining genuies and impostors for the fit
        genuines_X, genuines_Y = np.histogram(
            X[y == 1], bins=bins, density=True
        )
        # averaging the bins
        genuines_Y = 0.5 * (genuines_Y[1:] + genuines_Y[:-1])

        # gamma fit for impostor distribution
        impostors_popt, _ = curve_fit(
            beta_pdf, xdata=impostors_X, ydata=impostors_Y, p0=[1, 0.5]
        )
        impostors_a, impostors_b = impostors_popt

        # Fitting weibull for genuines and impostors
        genuines_popt, _ = curve_fit(
            beta_pdf, xdata=genuines_X, ydata=genuines_Y, p0=[1.0, 0.5]
        )
        genuines_a, genuines_b = genuines_popt

        self.gen_dist = partial(beta_pdf, a=genuines_a, b=genuines_b)
        self.imp_dist = partial(beta_pdf, a=impostors_a, b=impostors_b)
        return self

    def predict_proba(self, X):
        epsilon = 1e-10
        return +1 * (
            np.log10(self.gen_dist(X) + epsilon)
            - np.log10(self.imp_dist(X) + epsilon)
        )


class CategoricalCalibration(TransformerMixin, BaseEstimator):
    r"""
    Implements an adaptation of the Categorical Calibration defined in:

    `Mandasari, Miranti Indar, et al. "Score calibration in face recognition." Iet Biometrics 3.4 (2014): 246-256.`

    In such a work the calibration is defined as::
      :math:`s = \sum_{i=0}^{N} (\text{calibrator}_i)(X)`


    The category calibration is implemented in the tails of the score distributions in this implementation.
    For the impostor score distribution, the tail is defined between :math:`q_3(x)` and :math:`q3(x)+\alpha * (q3(x)-q1(x))`, where :math:`q_n` represents the quantile and :math:`\alpha` represents an offset.
    For the genuines score distribution, the tail is defined between :math:`q_1(x)` and :math:`q1(x)-\alpha * (q3(x)-q1(x))`.

    In this implementation one calibrator per category is fitted at training time.
    At test time, the maximum of the calibrated scores is returned.



    Parameters
    ----------
        field_name: str
           Reference field name in the csv score file. E.g. race, gender, ..,

        field_values: list
           Possible values for `field_name`. E.g ['male', 'female'], ['black', 'white']

        score_selection_method: str
           Method to select the scores for fetting the calibration models.
             `median-q3`: It will select the scores from the median to q3 from the impostor scores (q1 to median for genuines)
             `q3-outlier`: It will select the scores from q3 to outlier (q3+1.5*IQD) from the impostor scores (q1 to outlier for genuines)
             `q1-median`:
             `all`: It will select all the scores. Default to `median-q3`

        reduction_function:
           Pointer to a function to reduce the scores. Default to `np.mean`

        fit_estimator: None
           Estimator used for calibrations. Default to `LLRCalibration`
    """

    def __init__(
        self,
        field_name,
        field_values,
        score_selection_method="all",
        reduction_function=np.mean,
        fit_estimator=None,
    ):
        self.field_name = field_name
        self.field_values = field_values
        self.quantile_factor = 1.5
        self.score_selection_method = score_selection_method
        self.reduction_function = reduction_function

        if fit_estimator is None:
            self.fit_estimator = LLRCalibration
        else:
            self.fit_estimator = fit_estimator

    def fit(self, input_score_file_name):
        """
        Fit the calibrator

        Parameters
        ----------

           input_score_file_name: str
              Reference score file used to fit the calibrator (E.g `scores-dev.csv`).


        """

        def impostor_threshold(impostor_scores):
            """
            score > Q3 + 1.5*IQR
            """
            q1 = np.quantile(impostor_scores, 0.25)
            q3 = np.quantile(impostor_scores, 0.75)

            if self.score_selection_method == "q3-outlier":
                outlier_zone = q3 + self.quantile_factor * (q3 - q1)
                return impostor_scores[
                    (impostor_scores > q3) & (impostor_scores <= outlier_zone)
                ]
            elif self.score_selection_method == "q1-median":
                median = np.median(impostor_scores)
                return impostor_scores[
                    (impostor_scores > q1) & (impostor_scores <= median)
                ]
            elif self.score_selection_method == "median-q3":
                median = np.median(impostor_scores)
                return impostor_scores[
                    (impostor_scores < q3) & (impostor_scores >= median)
                ]
            else:
                return impostor_scores

        def genuine_threshold(genuine_scores):
            """
            score <= Q3 - 1.5*IQR
            """
            q1 = np.quantile(genuine_scores, 0.25)
            q3 = np.quantile(genuine_scores, 0.75)

            if self.score_selection_method == "q3-outlier":
                outlier_zone = q1 - self.quantile_factor * (q3 - q1)
                return genuine_scores[
                    (genuine_scores < q1) & (genuine_scores >= outlier_zone)
                ]
            elif self.score_selection_method == "q1-median":
                median = np.median(genuine_scores)
                return genuine_scores[
                    (genuine_scores < q3) & (genuine_scores <= median)
                ]
            elif self.score_selection_method == "median-q3":
                median = np.median(genuine_scores)
                return genuine_scores[
                    (genuine_scores > q1) & (genuine_scores <= median)
                ]
            else:
                return genuine_scores

        impostors, genuines = get_split_dataframe(input_score_file_name)
        self._categorical_fitters = []

        def get_sigmoid_score(X, regressor):
            return expit(X * regressor.coef_ + regressor.intercept_).ravel()

        for value in self.field_values:
            # Filtering genunines and impostors per group
            impostors_per_group = (
                impostors[
                    (impostors[f"probe_{self.field_name}"] == value)
                    & (impostors[f"bio_ref_{self.field_name}"] == value)
                ]["score"]
                .compute()
                .to_numpy()
            )
            genuines_per_group = (
                genuines[(genuines[f"probe_{self.field_name}"] == value)][
                    "score"
                ]
                .compute()
                .to_numpy()
            )

            impostors_per_group = impostor_threshold(impostors_per_group)
            genuines_per_group = genuine_threshold(genuines_per_group)

            # Training the regressor
            y = np.hstack(
                (
                    np.zeros(len(impostors_per_group)),
                    np.ones(len(genuines_per_group)),
                )
            )
            X = np.expand_dims(
                np.hstack((impostors_per_group, genuines_per_group)), axis=1
            )
            fitter = self.fit_estimator().fit(X, y)
            self._categorical_fitters.append(fitter)

        return self

    def transform(self, input_scores, calibrated_scores):
        """
        Calibrates a score

        Parameters
        ----------

           input_scores: list
              Input score files to be calibrated

           calibrated_files: list
              Output score files

        """

        assert isinstance(input_scores, list) or isinstance(input_scores, tuple)
        assert isinstance(calibrated_scores, list) or isinstance(
            calibrated_scores, tuple
        )
        assert len(calibrated_scores) == len(input_scores)
        for file_name, output_file_name in zip(input_scores, calibrated_scores):
            # Fetching scores
            dataframe = dask.dataframe.read_csv(file_name)
            dataframe = dataframe.compute()
            X = dataframe["score"].to_numpy()

            calibrated_scores = np.vstack(
                [
                    fitter.predict_proba(X)
                    for fitter in self._categorical_fitters
                ]
            ).T
            calibrated_scores = self.reduction_function(
                calibrated_scores, axis=1
            )
            dataframe["score"] = calibrated_scores

            dataframe.to_csv(output_file_name, index=False)

        return calibrated_scores



"""Tests for bob.measure scripts"""

import shutil

import click
import numpy
import pkg_resources

from click.testing import CliRunner

from bob.bio.base.score import scores
from bob.bio.base.script import commands, compare_samples, sort, vuln_commands
from bob.io.base.testing_utils import assert_click_runner_result


def test_metrics():
    dev1 = pkg_resources.resource_filename(__name__, "data/dev-4col.txt")
    runner = CliRunner()
    result = runner.invoke(commands.metrics, [dev1])
    with runner.isolated_filesystem():
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)
    dev2 = pkg_resources.resource_filename(__name__, "data/dev-5col.txt")
    test1 = pkg_resources.resource_filename(__name__, "data/test-4col.txt")
    test2 = pkg_resources.resource_filename(__name__, "data/test-5col.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics, ["-e", dev1, test1, dev2, test2]
        )
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics,
            ["-e", "-l", "tmp", "-lg", "A,B", dev1, test1, dev2, test2],
        )
        assert_click_runner_result(result)
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics, ["-e", "-l", "tmp", dev1, test2]
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics,
            [
                "-e",
                "-l",
                "tmp",
                "-T",
                "0.1",
                "--criterion",
                "mindcf",
                "--cost",
                0.9,
                dev1,
                test2,
            ],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics,
            ["-l", "tmp", "--criterion", "mindcf", "--cost", 0.9, dev1],
        )
        assert_click_runner_result(result)

    # with runner.isolated_filesystem():
    #     result = runner.invoke(
    #         commands.metrics, ["-e", "--criterion", "cllr", dev1, test2]
    #     )
    #     assert_click_runner_result(result)

    # with runner.isolated_filesystem():
    #     result = runner.invoke(
    #         commands.metrics, ["-l", "tmp", "--criterion", "cllr", "--cost", 0.9, dev1]
    #     )
    #     assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics,
            ["-e", "--criterion", "rr", "-T", "0.1", dev1, test2],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics, ["-l", "tmp", "--criterion", "rr", dev1, dev2]
        )
        assert_click_runner_result(result)


def test_roc():
    """

    dev1 = pkg_resources.resource_filename('bob.bio.base.test',
                                           'data/dev-4col.txt')
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.roc, ['--output',
                                              'test.pdf',dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    dev2 = pkg_resources.resource_filename('bob.bio.base.test',
                                           'data/dev-5col.txt')
    test1 = pkg_resources.resource_filename('bob.bio.base.test',
                                            'data/test-4col.txt')
    test2 = pkg_resources.resource_filename('bob.bio.base.test',
                                            'data/test-5col.txt')
    with runner.isolated_filesystem():
        result = runner.invoke(commands.roc, ['--split', '--output',
                                              'test.pdf', '-S', '-ll',
                                              'lower-left', '-e',
                                              dev1, test1, dev2, test2])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


    with runner.isolated_filesystem():
        result = runner.invoke(commands.roc, ['--output',
                                              'test.pdf',
                                              '-e', '--legends', 'A,B',
                                              dev1, test1, dev2, test2])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    dev_nonorm = pkg_resources.resource_filename('bob.bio.base.test',
                                                 'data/scores-nonorm-dev')
    dev_ztnorm = pkg_resources.resource_filename('bob.bio.base.test',
                                                 'data/scores-ztnorm-dev')
    with runner.isolated_filesystem():
        result = runner.invoke(commands.roc, [
            '--min-far-value', '1e-6',
            '--lines-at', '1e-5',
            '-v', '--legends', 'A', '-e',
            dev_nonorm, dev_ztnorm
        ])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    """
    pass


def test_det():
    dev1 = pkg_resources.resource_filename(__name__, "data/dev-4col.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.det, [dev1, "-S"])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    dev2 = pkg_resources.resource_filename(__name__, "data/dev-5col.txt")
    test1 = pkg_resources.resource_filename(__name__, "data/test-4col.txt")
    test2 = pkg_resources.resource_filename(__name__, "data/test-5col.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.det,
            [
                "--split",
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "-e",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.det,
            ["--output", "test.pdf", "-e", dev1, test1, dev2, test2],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    dev_nonorm = pkg_resources.resource_filename(
        __name__, "data/scores-nonorm-dev"
    )
    dev_ztnorm = pkg_resources.resource_filename(
        __name__, "data/scores-ztnorm-dev"
    )
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.det,
            [
                "--min-far-value",
                "1e-6",
                "--lines-at",
                "1e-5",
                "-e",
                "-v",
                "--legends",
                "A",
                dev_nonorm,
                dev_ztnorm,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_epc():
    dev1 = pkg_resources.resource_filename(__name__, "data/dev-4col.txt")
    test1 = pkg_resources.resource_filename(__name__, "data/test-4col.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.epc, [dev1, test1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    dev2 = pkg_resources.resource_filename(__name__, "data/dev-4col.tar.gz")
    test2 = pkg_resources.resource_filename(__name__, "data/test-5col.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.epc,
            [
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "-S",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    dev_nonorm = pkg_resources.resource_filename(
        __name__, "data/scores-nonorm-dev"
    )
    dev_ztnorm = pkg_resources.resource_filename(
        __name__, "data/scores-ztnorm-dev"
    )

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.epc, ["-v", "--legends", "A", dev_nonorm, dev_ztnorm]
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_hist():
    dev1 = pkg_resources.resource_filename(__name__, "data/dev-4col.txt")
    dev2 = pkg_resources.resource_filename(__name__, "data/dev-5col.txt")
    test1 = pkg_resources.resource_filename(__name__, "data/test-4col.txt")
    test2 = pkg_resources.resource_filename(__name__, "data/test-5col.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.hist, [dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            [
                "--criterion",
                "min-hter",
                "--output",
                "HISTO.pdf",
                "-b",
                "30,auto",
                dev1,
                dev2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            [
                "--criterion",
                "eer",
                "--output",
                "HISTO.pdf",
                "-b",
                "30",
                "-e",
                "-ts",
                "A,B",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_cmc():
    dev1 = pkg_resources.resource_filename(__name__, "data/scores-cmc-5col.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.cmc, [dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    test1 = pkg_resources.resource_filename(
        __name__, "data/scores-cmc-4col.txt"
    )
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.cmc,
            [
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "-S",
                "-ts",
                "TA,TB",
                "-e",
                dev1,
                test1,
                dev1,
                test1,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    dev_nonorm = pkg_resources.resource_filename(
        __name__, "data/scores-nonorm-dev"
    )
    dev_ztnorm = pkg_resources.resource_filename(
        __name__, "data/scores-ztnorm-dev"
    )

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.cmc, ["-v", "--legends", "A", "-e", dev_nonorm, dev_ztnorm]
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_dir():
    dev1 = pkg_resources.resource_filename(
        __name__, "data/scores-nonorm-openset-dev"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.dir, [dev1, "--rank", 2])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    test1 = pkg_resources.resource_filename(
        __name__, "data/scores-nonorm-openset-dev"
    )
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.dir,
            [
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "-S",
                "--min-far-value",
                "1e-6",
                "-e",
                dev1,
                test1,
                dev1,
                test1,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_sort():
    def sorted_scores(score_lines):
        lines = []
        floats = []
        for line in score_lines:
            lines.append(line)
            floats.append(line[-1])
        sort_idx = numpy.argsort(floats)
        lines = [lines[i] for i in sort_idx]
        return lines

    dev1 = pkg_resources.resource_filename(__name__, "data/scores-nonorm-dev")
    runner = CliRunner()
    with runner.isolated_filesystem():
        # create a temporary sort file and sort it and check if it is sorted!

        path = "scores.txt"
        shutil.copy(dev1, path)

        result = runner.invoke(sort.sort, [path])
        assert_click_runner_result(result, exit_code=0)

        # load dev1 and sort it and compare to path
        dev1_sorted = list(sorted_scores(scores(dev1)))
        path_scores = list(scores(path))

        assert dev1_sorted == path_scores


def test_metrics_vuln():
    dev1 = pkg_resources.resource_filename(__name__, "data/vuln/scores-dev.csv")
    runner = CliRunner()
    result = runner.invoke(vuln_commands.metrics, [dev1])
    with runner.isolated_filesystem():
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)
    dev2 = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev-med.csv"
    )
    test1 = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    test2 = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval-med.csv"
    )
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.metrics, ["-e", dev1, test1, dev2, test2]
        )
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.metrics,
            ["-e", "-l", "tmp", "-lg", "A,B", dev1, test1, dev2, test2],
        )
        assert_click_runner_result(result)
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.metrics, ["-e", "-l", "tmp", dev1, test2]
        )
        assert_click_runner_result(result)


def test_det_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.det,
            ["-fnmr", "0.2", "-o", "DET.pdf", dev_file, eval_file],
        )
        assert_click_runner_result(result)


def test_fmr_iapmr_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.fmr_iapmr,
            [
                "--output",
                "FMRIAPMR.pdf",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)

        result = runner.invoke(
            vuln_commands.fmr_iapmr,
            [
                "--output",
                "FMRIAPMR.pdf",
                dev_file,
                eval_file,
                "-G",
                "-L",
                "1e-7,1,0,1",
            ],
        )
        assert_click_runner_result(result)


def test_hist_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.hist,
            [
                "--criterion",
                "eer",
                "--output",
                "HISTO.pdf",
                "-b",
                "30",
                "-ts",
                "A,B",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.hist,
            [
                "--criterion",
                "eer",
                "--output",
                "HISTO.pdf",
                "-b",
                "2,20,30",
                "-e",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)


def test_epc_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.epc,
            [
                "--output",
                "epc.pdf",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)

        result = runner.invoke(
            vuln_commands.epc,
            [
                "--output",
                "epc.pdf",
                "-I",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)


def test_epsc_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.epsc,
            [
                "--output",
                "epsc.pdf",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)

        result = runner.invoke(
            vuln_commands.epsc,
            [
                "--output",
                "epsc.pdf",
                "-I",
                "-fp",
                "0.1,0.3",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)


def test_epsc_3D_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.epsc,
            [
                "--output",
                "epsc.pdf",
                "-D",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)

        result = runner.invoke(
            vuln_commands.epsc,
            [
                "--output",
                "epsc.pdf",
                "-D",
                "-I",
                "--no-wer",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)


def test_evaluate_vuln():
    dev_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    eval_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            vuln_commands.evaluate,
            [
                "-e",
                "-o",
                "evaluate_vuln.pdf",
                dev_file,
                eval_file,
            ],
        )
        assert_click_runner_result(result)


def test_compare_samples():
    sample_1_path = pkg_resources.resource_filename(
        __name__, "data/dummy_samples_1.hdf5"
    )
    sample_2_path = pkg_resources.resource_filename(
        __name__, "data/dummy_samples_2.hdf5"
    )
    runner = CliRunner()
    from bob.bio.base.config.dummy.pipeline import pipeline

    with runner.isolated_filesystem():
        result = runner.invoke(
            compare_samples.compare_samples,
            [
                "--pipeline",
                pipeline,
                sample_1_path,
                sample_2_path,
            ],
        )
        assert_click_runner_result(result)



import numpy as np

from scipy.spatial.distance import cdist

from ..pipelines import BioAlgorithm


class Distance(BioAlgorithm):
    """A distance algorithm to compare feature vectors.
    Many biometric algorithms are based on comparing feature vectors that
    are usually extracted by using deep neural networks.
    The most common distance function is the cosine similarity, which is
    the default in this class.
    """

    def __init__(
        self,
        distance_function="cosine",
        factor=-1,
        average_on_enroll=True,
        average_probes=False,
        probes_score_fusion="max",
        enrolls_score_fusion="max",
        **kwargs,
    ):
        """
        Parameters
        ----------
        distance_function : str or :py:class:`function`, optional
            function to be used to measure the distance of probe and model
            features compatible with :any:`scipy.spatial.distance.cdist`. If the
            function exists in scipy.spatial.distance, provide its string name
            as scipy will run an optimized version.

        factor : float
            A coefficient which is multiplied to distance (after
            distance_function) to find score between probe and model features.
            In bob.bio.base, the scores should be similarity scores (higher
            score for a genuine pair) so use this factor to make sure you are
            using similarity scores.

        average_on_enroll : bool
            Some database protocols contain multiple samples (e.g. face images)
            to create one enrollment template. This option is useful in case of
            those databases. If True, the algorithm will average the enroll
            features to create a single template. If False, the algorithm will
            use the enroll features as is and will compare the probe template
            against all features. The final score will be computed based on the
            ``enrolls_score_fusion`` option.

        average_probes : bool
            Some database protocols contain multiple samples (e.g. face images)
            to create one probe template. This option is useful in case of those
            databases. If True, the algorithm will average the probe features to
            create a single template. If False, the algorithm will use the probe
            features as is and will compare the enrollment template against all
            features. The final score will be computed based on the
            ``probes_score_fusion`` option.

        probes_score_fusion : str
            How to fuse the scores of the probes if average_probes is False and
            the database contains multiple probe samples.

        enrolls_score_fusion : str
            How to fuse the scores of the enrolls if average_on_enroll is False
            and the database contains multiple enroll samples.
        """
        super().__init__(
            probes_score_fusion=probes_score_fusion,
            enrolls_score_fusion=enrolls_score_fusion,
            **kwargs,
        )
        self.distance_function = distance_function
        self.factor = factor
        self.average_on_enroll = average_on_enroll
        self.average_probes = average_probes

    def create_templates(self, list_of_feature_sets, enroll):
        """Creates templates from the given feature sets.
        Will make sure the features are 2 dimensional before creating templates.
        Will average features over samples if ``average_on_enroll`` is True or
        ``average_probes`` is True.
        """
        list_of_feature_sets = [
            self._make_2d(data) for data in list_of_feature_sets
        ]
        # shape of list_of_feature_sets is Nx?xD
        if (enroll and self.average_on_enroll) or (
            not enroll and self.average_probes
        ):
            # we cannot call np.mean(list_of_feature_sets, axis=1) because the size of
            # axis 1 is diffent for each feature set.
            # output will be NxD
            return np.array(
                [np.mean(feat, axis=0) for feat in list_of_feature_sets]
            )
        # output shape is Nx?xD
        return list_of_feature_sets

    def _make_2d(self, X):
        """Makes sure that the features are 2 dimensional before creating enroll
        and probe templates.

        For instance, when the source is `VideoLikeContainer` the input of
        ``create_templates`` is [`VideoLikeContainer`, ....]. The concatenation
        of them makes and array of `ZxNxD`. Hence we need to stack them in `Z`.
        """
        if not len(X):
            return [[]]
        if X[0].ndim == 2:
            X = np.vstack(X)
        return np.atleast_2d(X)

    def compare(self, enroll_templates, probe_templates):
        """Compares the probe templates to the enroll templates.

        Depending on the ``average_on_enroll`` and ``average_probes`` options,
        the templates have different shapes.
        """
        # returns scores NxM where N is the number of enroll templates and M is
        # the number of probe templates
        if self.average_on_enroll and self.average_probes:
            # enroll_templates is NxD
            enroll_templates = np.asarray(enroll_templates)
            # probe_templates is MxD
            probe_templates = np.asarray(probe_templates)
            return self.factor * cdist(
                enroll_templates, probe_templates, self.distance_function
            )
        elif self.average_on_enroll:
            # enroll_templates is NxD
            enroll_templates = np.asarray(enroll_templates)
            # probe_templates is Mx?xD
            scores = []
            for probe in probe_templates:
                s = self.factor * cdist(
                    enroll_templates, probe, self.distance_function
                )
                # s is Nx?, we want s to be N
                s = self.fuse_probe_scores(s, axis=1)
                scores.append(s)
            return np.array(scores).T
        elif self.average_probes:
            # enroll_templates is Nx?xD
            # probe_templates is MxD
            probe_templates = np.asarray(probe_templates)
            scores = []
            for enroll in enroll_templates:
                s = self.factor * cdist(
                    enroll, probe_templates, self.distance_function
                )
                # s is ?xM, we want s to be M
                s = self.fuse_enroll_scores(s, axis=0)
                scores.append(s)
            return np.array(scores)
        else:
            # enroll_templates is Nx?1xD
            # probe_templates is Mx?2xD
            scores = []
            for enroll in enroll_templates:
                scores.append([])
                for probe in probe_templates:
                    s = self.factor * cdist(
                        enroll, probe, self.distance_function
                    )
                    # s is ?1x?2, we want s to be scalar
                    s = self.fuse_probe_scores(s, axis=1)
                    s = self.fuse_enroll_scores(s, axis=0)
                    scores[-1].append(s)
            return np.array(scores)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


import logging

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from bob.pipelines import is_pipeline_wrapped
from bob.pipelines.distributed import (
    VALID_DASK_CLIENT_STRINGS,
    dask_get_partition_size,
)

logger = logging.getLogger(__name__)
from bob.pipelines.wrappers import CheckpointWrapper, DaskWrapper, wrap

EPILOG = """\b
Command line examples\n
---------------------\n


Follow below an example on how to extract arcface features from a database:\n

`bob bio transform my_database iresnet100 -vv`\n\n

To "dask" the execution of the pipeline, you can use the `--dask-client` option.\n
In the example below we show how to use the `--dask-client` option to start a dask cluster on SGE.\n

`bob bio transform my_database iresnet100 --dask-client sge -vv`\n\n

\b

Creating my own transformer\n
---------------------------\n

This command accepts configuration file as input.\n
For example, if you desire to customize your transformer, you can use the following configuration file:\n
\b\b


```py\n
from sklearn.base import BaseEstimator, TransformerMixin \n
from sklearn.pipeline import make_pipeline  \n
from bob.pipelines import wrap \n

class MyTransformer(TransformerMixin, BaseEstimator): \n
    def _more_tags(self): \n
        return {"requires_fit": False} \n

    def transform(self, X): \n
        # do something \n
        return X \n

transformer = wrap(["sample"],make_pipeline(MyTransformer())) \n
```

Then, you can use above configuration file to run the command:

\b
`bob bio pipelines transform my_database my_transformer.py --dask-client sge -vv`

\b\b

Leveraging from FunctionTransformer\n
-----------------------------------\n

You can also benefit from `FunctionTransformer` to create a transformer

```py \n
from sklearn.preprocessing import FunctionTransformer \n
from sklearn.pipeline import make_pipeline \n
from bob.pipelines import wrap \n

\b

def my_transformer(X): \n
    # do something \n
    return X \n


transformer = wrap(["sample"],make_pipeline(FunctionTransformer(my_transformer))) \n
```

Then, you can use above configuration file to run the command:\n

`bob bio pipelines transform my_database my_transformer.py --dask-client sge -vv`


\b\b
"""


@click.command(
    name="transform",
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog=EPILOG,
)
@click.option(
    "--transformer",
    "-t",
    required=True,
    entry_point_group="bob.bio.transformer",
    help="A scikit-learn Pipeline containing the set of transformations",
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    entry_point_group="bob.bio.database",
    required=True,
    help="Biometric Database connector (class that implements the methods: `background_model_samples`, `references` and `probes`)",
    cls=ResourceOption,
)
@click.option(
    "--dask-client",
    "-l",
    entry_point_group="dask.client",
    string_exceptions=VALID_DASK_CLIENT_STRINGS,
    default="single-threaded",
    help="Dask client for the execution of the pipeline.",
    cls=ResourceOption,
)
@click.option(
    "-c",
    "--checkpoint-dir",
    show_default=True,
    default="./checkpoints",
    help="Name of output directory where the checkpoints will be saved.",
    cls=ResourceOption,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set",
    cls=ResourceOption,
)
@click.option(
    "--dask-partition-size",
    "-s",
    help="If using Dask, this option defines the size of each dask.bag.partition."
    "Use this option if the current heuristic that sets this value doesn't suit your experiment."
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@verbosity_option(cls=ResourceOption, logger=logger)
def pipeline_transform(
    transformer,
    database,
    dask_client,
    checkpoint_dir,
    force,
    dask_partition_size,
    **kwargs,
):
    """
    This CLI command will execute a pipeline (a scikit learn Pipeline)  on a given database.

    This command can be used, for example, to extract features, face-crops, preprocess audio files and so on.

    """

    logger.info(f"Transforming samples from {database}")

    # save_func = bob.io.base.save

    # Idiap SETUP. This avoids having directories with more than 1000 files/directories
    hash_fn = database.hash_fn if hasattr(database, "hash_fn") else None

    # If NONE of the items are checkpointed, we checkpoint them all
    if not any(is_pipeline_wrapped(transformer, CheckpointWrapper)):
        logger.info("Checkpointing it")
        transformer = wrap(
            ["checkpoint"],
            transformer,
            features_dir=checkpoint_dir,
            hash_fn=hash_fn,
            force=force,
        )
    else:
        # If there is only one item that is checkpointed, we don't need to wrap the pipeline
        logger.warning(
            f"{transformer}"
            f"The pipeline contains elements that are already checkpointed."
            "Hence, we are not checkpointing them again."
        )

    # Fetching all samples
    samples = database.all_samples()

    # The number of dasked elements has to be the number of
    # elements in the pipeline - 1 (the ToDaskBag doesn't count)
    dasked_elements = is_pipeline_wrapped(transformer, DaskWrapper)

    if any(dasked_elements):
        logger.warning(
            "The pipeline is already dasked, hence, we are not dasking it again."
        )
    else:
        if not isinstance(dask_client, str):
            dask_partition_size = (
                dask_get_partition_size(
                    dask_client.cluster, len(samples), lower_bound=200
                )
                if dask_partition_size is None
                else dask_partition_size
            )

        logger.info(
            f"Dask wrapping it with partition size {dask_partition_size}"
        )
        transformer = wrap(
            ["dask"], transformer, partition_size=dask_partition_size
        )

    transformer.transform(samples).compute(
        scheduler="single-threaded" if dask_client is None else dask_client
    )

    logger.info("Transformation finished !")



import bz2
import glob
import hashlib
import io
import logging
import tarfile
import warnings
import zipfile

from fnmatch import fnmatch
from os import PathLike
from pathlib import Path
from typing import IO, Any, Callable, TextIO, Union

import requests

from clapper.rc import UserDefaults

logger = logging.getLogger(__name__)


def _get_local_data_directory() -> Path:
    user_config = UserDefaults("bobrc.toml")
    return Path(
        user_config.get("bob_data_dir", default=Path.home() / "bob_data")
    )


def _path_and_subdir(
    archive_path: Union[str, PathLike[str]],
) -> tuple[Path, Union[Path, None]]:
    """Splits an archive's path from a sub directory (separated by ``:``)."""
    archive_path_str = Path(archive_path).as_posix()
    if ":" in archive_path_str:
        archive, sub_dir = archive_path_str.rsplit(":", 1)
        return Path(archive), Path(sub_dir)
    return Path(archive_path), None


def _is_bz2(path: Union[str, PathLike[str]]) -> bool:
    try:
        with bz2.BZ2File(path) as f:
            f.read(1024)
        return True
    except (OSError, EOFError):
        return False


def is_archive(path: Union[str, PathLike[str]]) -> bool:
    """Returns whether the path points in an archive.

    Any path pointing to a valid tar or zip archive or to a valid bz2
    file will return ``True``.
    """
    archive = _path_and_subdir(path)[0]
    try:
        return any(
            tester(_path_and_subdir(archive)[0])
            for tester in (tarfile.is_tarfile, zipfile.is_zipfile, _is_bz2)
        )
    except (FileNotFoundError, IsADirectoryError):
        return False


def search_in_archive_and_open(
    search_pattern: str,
    archive_path: Union[str, PathLike[str]],
    inner_dir: Union[str, PathLike[str], None] = None,
    open_as_binary: bool = False,
) -> Union[IO[bytes], TextIO, None]:
    """Returns a read-only stream of a file matching a pattern in an archive.

    Wildcards (``*``, ``?``, and ``**``) are supported (using
    :meth:`pathlib.Path.glob`).

    The first matching file will be open and returned.

    examples:

    .. code-block: text

        archive.tar.gz
            + subdir1
            |   + file1.txt
            |   + file2.txt
            |
            + subdir2
                + file1.txt

    ``search_and_open("archive.tar.gz", "file1.txt")``
    opens``archive.tar.gz/subdir1/file1.txt``

    ``search_and_open("archive.tar.gz:subdir2", "file1.txt")``
    opens ``archive.tar.gz/subdir2/file1.txt``

    ``search_and_open("archive.tar.gz", "*.txt")``
    opens ``archive.tar.gz/subdir1/file1.txt``


    Parameters
    ----------
    archive_path
        The ``.tar.gz`` archive file containing the wanted file. To match
        ``search_pattern`` in a sub path in that archive, append the sub path
        to ``archive_path`` with a ``:`` (e.g.
        ``/path/to/archive.tar.gz:sub/dir/``).
    search_pattern
        A string to match to the file. Wildcards are supported (Unix pattern
        matching).

    Returns
    -------
    io.TextIOBase or io.BytesIO
        A read-only file stream.
    """

    archive_path = Path(archive_path)

    if inner_dir is None:
        archive_path, inner_dir = _path_and_subdir(archive_path)

    if inner_dir is not None:
        pattern = (Path("/") / inner_dir / search_pattern).as_posix()
    else:
        pattern = (Path("/") / search_pattern).as_posix()

    if ".tar" in archive_path.suffixes:
        tar_arch = tarfile.open(archive_path)  # TODO File not closed
        for member in tar_arch:
            if member.isfile() and fnmatch("/" + member.name, pattern):
                break
        else:
            logger.debug(
                f"No file matching '{pattern}' were found in '{archive_path}'."
            )
            return None

        if open_as_binary:
            return tar_arch.extractfile(member)
        return io.TextIOWrapper(tar_arch.extractfile(member), encoding="utf-8")

    elif archive_path.suffix == ".zip":
        zip_arch = zipfile.ZipFile(archive_path)
        for name in zip_arch.namelist():
            if fnmatch("/" + name, pattern):
                break
        else:
            logger.debug(
                f"No file matching '{pattern}' were found in '{archive_path}'."
            )
        return zip_arch.open(name)

    raise ValueError(
        f"Unknown file extension '{''.join(archive_path.suffixes)}'"
    )


def list_dir_in_archive(
    archive_path: Union[str, PathLike[str]],
    inner_dir: Union[str, PathLike[str], None] = None,
    show_dirs: bool = True,
    show_files: bool = True,
) -> list[Path]:
    """Returns a list of all the elements in an archive or inner directory.

    Parameters
    ----------
    archive_path
        A path to an archive, or an inner directory of an archive (appended
        with a ``:``).
    inner_dir
        A path inside the archive with its root at the archive's root.
    show_dirs
        Returns directories.
    show_files
        Returns files.
    """

    archive_path, arch_inner_dir = _path_and_subdir(archive_path)
    inner_dir = Path(inner_dir or arch_inner_dir or Path("."))

    results = []
    # Read the archive info and iterate over the paths. Return the ones we want.
    if ".tar" in archive_path.suffixes:
        with tarfile.open(archive_path) as arch:
            for info in arch.getmembers():
                path = Path(info.name)
                if path.parent != inner_dir:
                    continue
                if info.isdir() and show_dirs:
                    results.append(Path("/") / path)
                if info.isfile() and show_files:
                    results.append(Path("/") / path)
    elif archive_path.suffix == ".zip":
        with zipfile.ZipFile(archive_path) as arch:
            for zip_info in arch.infolist():
                zip_path = zipfile.Path(archive_path, zip_info.filename)
                if Path(zip_info.filename).parent != inner_dir:
                    continue
                if zip_path.is_dir() and show_dirs:
                    results.append(Path("/") / zip_info.filename)
                if not zip_path.is_dir() and show_files:
                    results.append(Path("/") / zip_info.filename)
    elif archive_path.suffix == ".bz2":
        if inner_dir != Path("."):
            raise ValueError(
                ".bz2 files don't have an inner structure (tried to access "
                f"'{archive_path}:{inner_dir}')."
            )
        results.extend([Path(archive_path.stem)] if show_files else [])
    else:
        raise ValueError(
            f"Unsupported archive extension '{''.join(archive_path.suffixes)}'."
        )
    return sorted(results)  # Fixes inconsistent file ordering across platforms


def extract_archive(
    archive_path: Union[str, PathLike[str]],
    inner_path: Union[str, PathLike[str], None] = None,
    destination: Union[str, PathLike[str], None] = None,
) -> Path:
    """Extract an archive and returns the location of the extracted data.

    Supports ``.zip``, ``.tar.gz``, ``.tar.bz2``, ``.tar.tgz``, and
    ``.tar.tbz2`` archives.
    Can also extract ``.bz2`` compressed files.

    Parameters
    ----------
    archive_path
        The compressed archive location. Pointing to a location inside a
        tarball can be achieved by appending ``:`` and the desired member to
        extract.
    inner_path
        A path with its root at the root of the archive file pointing to a
        specific file to extract.
    destination
        The desired location of the extracted file or directory. If not
        provided, the archive will be extracted where it stands (the parent of
        ``archive_path``).

    Returns
    -------
    pathlib.Path
        The extracted file or directory location.
        As an archive can contain any number of members, the parent directory
        is returned (where the archive content is extracted).

    Raises
    ------
    ValueError
        When ``archive_path`` does not point to a file with a known extension.
    """

    archive_path, arch_inner_dir = _path_and_subdir(archive_path)
    sub_dir = inner_path or arch_inner_dir

    if destination is None:
        destination = archive_path.parent

    if ".tar" in archive_path.suffixes:
        with tarfile.open(archive_path, mode="r") as arch:
            if sub_dir is None:
                arch.extractall(destination)
            else:
                arch.extract(Path(sub_dir).as_posix(), destination)
    elif ".zip" == archive_path.suffix:
        with zipfile.ZipFile(archive_path) as arch:
            if sub_dir is None:
                arch.extractall(destination)
            else:
                arch.extract(Path(sub_dir).as_posix(), destination)
    elif ".bz2" == archive_path.suffix:
        if sub_dir is not None:
            warnings.warn(
                f"Ignored sub directory ({sub_dir}). Not supported for `.bz2` files.",
                RuntimeWarning,
            )
        extracted_file = destination / Path(archive_path.stem)
        with bz2.BZ2File(archive_path) as arch, extracted_file.open(
            "wb"
        ) as dest:
            dest.write(arch.read())
    else:
        raise ValueError(
            f"Unknown file extension: {''.join(archive_path.suffixes)}"
        )
    return Path(destination)


def search_and_open(
    search_pattern: str,
    base_dir: Union[PathLike, None] = None,
    sub_dir: Union[PathLike, None] = None,
    open_as_binary: bool = False,
    **kwargs,
) -> Union[IO[bytes], TextIO, None]:
    """Searches for a matching file recursively in a directory.

    If ``base_dir`` points to an archive, the pattern will be searched inside that
    archive.

    Wildcards (``*``, ``?``, and ``**``) are supported (using
    :meth:`pathlib.Path.glob`).

    Parameters
    ----------
    search_pattern
        A string containing the wanted path pattern of the file to open.
        Supports ``fnmatch`` notation (``*``, ``**``, and ``?``).
    base_dir
        A path to a directory to search into. By default, will use the
        ``data_path`` user configuration.
    sub_dir
        A sub directory of ``base_dir`` to search into instead. Useful when
        using the default value of ``base_dir`` but still wanting to use a
        sub directory in there.
    open_as_binary
        Will open the file as a binary stream instead of a text file.

    Returns
    -------
    IO
        A read-only open file stream.
    """

    if base_dir is None:
        base_dir = _get_local_data_directory()

    if is_archive(base_dir):
        return search_in_archive_and_open(
            search_pattern=search_pattern,
            archive_path=base_dir,
            inner_dir=sub_dir,  # TODO not ok with config data_path / subdir
            **kwargs,
        )

    # If the input is local
    base_dir = Path(base_dir)
    final_dir = base_dir / sub_dir if sub_dir else base_dir
    if final_dir.is_dir():
        # we prepend './' to search_pattern because it might start with '/'
        pattern = final_dir / "**" / f"./{search_pattern}"
        for path in glob.iglob(pattern.as_posix(), recursive=True):
            if not Path(path).is_file():
                continue
            return open(path, mode="rb" if open_as_binary else "rt")
        raise FileNotFoundError(
            f"Unable to locate and open a file that matches '{pattern}' in "
            f"'{final_dir}'."
        )

    return open(final_dir, mode="rb" if open_as_binary else "rt")


def list_dir(
    base_directory: PathLike,
    sub_directory: Union[PathLike, None] = None,
    show_files: bool = True,
    show_dirs: bool = True,
) -> list[Path]:
    """Lists all directories and/or files in a directory (non-recursively)."""
    base_directory = Path(base_directory)

    if is_archive(base_directory):
        return list_dir_in_archive(
            archive_path=base_directory,
            inner_dir=sub_directory,
            show_dirs=show_dirs,
            show_files=show_files,
        )

    # Not an archive
    final_directory = (
        base_directory
        if sub_directory is None
        else base_directory / sub_directory
    )
    glob = list(final_directory.glob("*"))
    if not show_dirs:
        glob = [g for g in glob if not g.is_dir()]
    if not show_files:
        glob = [g for g in glob if not g.is_file()]
    return glob


def md5_hash(readable: Any, chunk_size: int = 65535) -> str:
    """Computes the md5 hash of any object with a read method."""
    hasher = hashlib.md5()
    for chunk in iter(lambda: readable.read(chunk_size), b""):
        hasher.update(chunk)
    return hasher.hexdigest()


def sha256_hash(readable: Any, chunk_size: int = 65535) -> str:
    """Computes the SHA256 hash of any object with a read method."""
    hasher = hashlib.sha256()
    for chunk in iter(lambda: readable.read(chunk_size), b""):
        hasher.update(chunk)
    return hasher.hexdigest()


def verify_file(
    file_path: Union[str, PathLike[str]],
    file_hash: str,
    hash_fct: Callable[[Any, int], str] = sha256_hash,
    full_match: bool = False,
) -> bool:
    """Returns True if the file computed hash corresponds to `file_hash`.

    For comfort, we allow ``file_hash`` to match with the first
    characters of the digest, allowing storing only e.g. the first 8
    char.

    Parameters
    ----------
    file_path
        The path to the file needing verification.
    file_hash
        The expected file hash digest.
    hash_fct
        A function taking a path and returning a digest. Defaults to SHA256.
    full_match
        If set to False, allows ``file_hash`` to match the first characters of
        the files digest (this allows storing e.g. 8 chars of a digest instead
        of the whole 64 characters of SHA256, and still matching.)
    """
    file_path = Path(file_path)
    with file_path.open("rb") as f:
        digest = hash_fct(f, 65535)
    return digest == file_hash if full_match else digest.startswith(file_hash)


def compute_crc(
    file_path: Union[str, PathLike[str]],
    hash_fct: Callable[[Any, int], str] = sha256_hash,
) -> str:
    """Returns the CRC of a file."""
    file_path = Path(file_path)
    with file_path.open("rb") as f:
        return hash_fct(f, 65535)


def _infer_filename_from_urls(urls=Union[list[str], str]) -> str:
    """Retrieves the remote filename from the URLs.

    Parameters
    ----------
    urls
        One or multiple URLs pointing to files with the same name.

    Returns
    -------
    The remote file name.

    Raises
    ------
    ValueError
        When urls point to files with different names.
    """
    if isinstance(urls, str):
        return urls.split("/")[-1]

    # Check that all urls point to the same file name
    names = [u.split("/")[-1] for u in urls]
    if not all(n == names[0] for n in names):
        raise ValueError(
            f"Cannot infer file name when urls point to different files ({names=})."
        )
    return urls[0].split("/")[-1]


def download_file(
    urls: Union[list[str], str],
    destination_directory: Union[str, PathLike[str], None] = None,
    destination_sub_directory: Union[str, None] = None,
    destination_filename: Union[str, None] = None,
    checksum: Union[str, None] = None,
    checksum_fct: Callable[[Any, int], str] = sha256_hash,
    force: bool = True,
    extract: bool = False,
    makedirs: bool = True,
) -> Path:
    """Downloads a remote file locally.

    This will overwrite any existing file with the same name.

    Parameters
    ----------
    urls
        The remote location of the server. If multiple addresses are given, we will try
        to download from them in order until one succeeds.
    destination_directory
        A path to a local directory where the file will be saved. If omitted, the file
        will be saved in the folder pointed by the ``wdr.local_directory`` key in the
        user configuration.
    destination_sub_directory
        An additional layer added to the destination directory (useful when using
        ``destination_directory=None``).
    destination_filename
        The final name of the local file. If omitted, the file will keep the name of
        the remote file.
    checksum
        When provided, will compute the file's checksum and compare to this.
    force
        Re-download and overwrite any existing file with the same name.
    extract
        Extract an archive or zip file next to the downloaded file.
        If this is set, the parent directory path will be returned.
    makedirs
        Automatically make the parent directories of the new local file.

    Returns
    -------
    The path to the new local file (or the parent directory if ``extract`` is True).

    Raises
    ------
    RuntimeError
        When the URLs provided are all invalid.
    ValueError
        When ``destination_filename`` is omitted and URLs point to files with different
        names.
        When the checksum of the file does not correspond to the provided ``checksum``.
    """

    if destination_filename is None:
        destination_filename = _infer_filename_from_urls(urls=urls)

    if destination_directory is None:
        destination_directory = _get_local_data_directory()

    destination_directory = Path(destination_directory)

    if destination_sub_directory is not None:
        destination_directory = (
            destination_directory / destination_sub_directory
        )

    local_file = destination_directory / destination_filename
    needs_download = True

    if not force and local_file.is_file():
        logger.info(
            f"File {local_file} already exists, skipping download ({force=})."
        )
        needs_download = False

    if needs_download:
        if isinstance(urls, str):
            urls = [urls]

        for tries, url in enumerate(urls):
            logger.debug(f"Retrieving file from '{url}'.")
            try:
                response = requests.get(url=url, timeout=10)
            except requests.exceptions.ConnectionError as e:
                if tries < len(urls) - 1:
                    logger.info(
                        f"Could not connect to {url}. Trying other URLs."
                    )
                logger.debug(e)
                continue

            logger.debug(
                f"http response: '{response.status_code}: {response.reason}'."
            )

            if response.ok:
                logger.debug(f"Got file from {url}.")
                break
            elif tries < len(urls) - 1:
                logger.info(
                    f"Failed to get file from {url}, trying other URLs."
                )
                logger.debug(f"requests.response was\n{response}")
        else:
            raise RuntimeError(
                f"Could not retrieve file from any of the provided URLs! ({urls=})"
            )

        if makedirs:
            local_file.parent.mkdir(parents=True, exist_ok=True)

        with local_file.open("wb") as f:
            f.write(response.content)

    if checksum is not None:
        if not verify_file(local_file, checksum, hash_fct=checksum_fct):
            if not needs_download:
                raise ValueError(
                    f"The local file hash does not correspond to '{checksum}' "
                    f"and {force=} prevents overwriting."
                )
            raise ValueError(
                "The downloaded file hash ('"
                f"{compute_crc(local_file, hash_fct=checksum_fct)}') does not "
                f"correspond to '{checksum}'."
            )

    if extract:
        local_file = extract_archive(local_file)

    return local_file



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import os
import tempfile

import numpy as np

import bob.io.base

from bob.bio.base.extractor import Extractor
from bob.bio.base.preprocessor import Preprocessor
from bob.bio.base.transformers import (
    ExtractorTransformer,
    PreprocessorTransformer,
)
from bob.pipelines import CheckpointWrapper, Sample, SampleWrapper


class FakePreprocessor(Preprocessor):
    def __call__(self, data, annotations=None):
        return data + annotations


class FakeExtractor(Extractor):
    def __call__(self, data):
        return data.flatten()[0:10]  # Selecting the first 10 features


class FakeExtractorFittable(Extractor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.requires_training = True
        self.model = None

    def __call__(self, data, metadata=None):
        model = self.model
        return data @ model

    def train(self, training_data, extractor_file):
        self.model = np.vstack(training_data)
        bob.io.base.save(self.model, extractor_file)


def generate_samples(
    n_subjects, n_samples_per_subject, shape=(2, 2), annotations=1
):
    """
    Simple sample generator that generates a certain number of samples per
    subject, whose data is np.zeros + subject index
    """

    samples = []
    for i in range(n_subjects):
        data = np.zeros(shape) + i
        for j in range(n_samples_per_subject):
            samples += [
                Sample(
                    data,
                    subject=str(i),
                    key=str(i * n_subjects + j),
                    annotations=annotations,
                )
            ]
    return samples


def assert_sample(transformed_sample, oracle):
    return np.alltrue(
        [np.allclose(ts.data, o) for ts, o in zip(transformed_sample, oracle)]
    )


def assert_checkpoints(transformed_sample, dir_name):
    return np.alltrue(
        [
            os.path.exists(os.path.join(dir_name, ts.key + ".h5"))
            for ts in transformed_sample
        ]
    )


def test_preprocessor():
    preprocessor = FakePreprocessor()
    preprocessor_transformer = PreprocessorTransformer(preprocessor)

    # Testing sample
    transform_extra_arguments = [("annotations", "annotations")]
    sample_transformer = SampleWrapper(
        preprocessor_transformer, transform_extra_arguments
    )

    data = np.zeros((2, 2))
    oracle = [np.ones((2, 2))]
    annotations = 1
    sample = [Sample(data, key="1", annotations=annotations)]
    transformed_sample = sample_transformer.transform(sample)

    assert assert_sample(transformed_sample, oracle)

    # Testing checkpoint
    with tempfile.TemporaryDirectory() as dir_name:
        checkpointing_transformer = CheckpointWrapper(
            sample_transformer,
            features_dir=dir_name,
            load_func=preprocessor.read_data,
            save_func=preprocessor.write_data,
        )
        transformed_sample = checkpointing_transformer.transform(sample)

        assert assert_sample(transformed_sample, oracle)
        assert assert_checkpoints(transformed_sample, dir_name)


def test_extractor():
    extractor = FakeExtractor()
    extractor_transformer = ExtractorTransformer(extractor)

    # Testing sample
    sample_transformer = SampleWrapper(extractor_transformer)

    data = np.zeros((2, 2))
    oracle = [np.zeros((1, 4))]
    sample = [Sample(data, key="1")]
    transformed_sample = sample_transformer.transform(sample)

    assert assert_sample(transformed_sample, oracle)

    # Testing checkpoint
    with tempfile.TemporaryDirectory() as dir_name:
        checkpointing_transformer = CheckpointWrapper(
            sample_transformer,
            features_dir=dir_name,
            load_func=extractor.read_feature,
            save_func=extractor.write_feature,
        )
        transformed_sample = checkpointing_transformer.transform(sample)

        assert assert_sample(transformed_sample, oracle)
        assert assert_checkpoints(transformed_sample, dir_name)


def test_extractor_fittable():
    with tempfile.TemporaryDirectory() as dir_name:
        extractor_file = os.path.join(dir_name, "Extractor.hdf5")
        extractor = FakeExtractorFittable()
        extractor_transformer = ExtractorTransformer(
            extractor, model_path=extractor_file
        )

        # Testing sample
        sample_transformer = SampleWrapper(extractor_transformer)
        # Fitting
        training_data = np.arange(4).reshape(2, 2)
        training_samples = [Sample(training_data, key="1")]
        sample_transformer = sample_transformer.fit(training_samples)

        test_data = [np.zeros((2, 2)), np.ones((2, 2))]
        oracle = [np.zeros((2, 2)), np.ones((2, 2)) @ training_data]
        test_sample = [Sample(d, key=str(i)) for i, d in enumerate(test_data)]

        transformed_sample = sample_transformer.transform(test_sample)
        assert assert_sample(transformed_sample, oracle)

        # Testing checkpoint
        checkpointing_transformer = CheckpointWrapper(
            sample_transformer,
            features_dir=dir_name,
            load_func=extractor.read_feature,
            save_func=extractor.write_feature,
        )
        transformed_sample = checkpointing_transformer.transform(test_sample)
        assert assert_sample(transformed_sample, oracle)
        assert assert_checkpoints(transformed_sample, dir_name)



#!/usr/bin/env python
# @author: Yannick Dayer <yannick.dayer@idiap.ch>
# @date: Fri 19 Aug 2022 14:37:01 UTC+02

import glob
import os
import tempfile

import pytest

from click.testing import CliRunner
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline

from bob.bio.base.pipelines.entry_points import execute_pipeline_train
from bob.bio.base.script.pipeline_train import (
    pipeline_train as pipeline_train_cli,
)
from bob.bio.base.wrappers import wrap_bob_legacy
from bob.io.base.testing_utils import assert_click_runner_result
from bob.pipelines import wrap
from tests.test_pipeline_simple import DummyDatabase
from tests.test_transformers import FakeExtractor, FakePreprocessor


class FittableTransformer(BaseEstimator):
    def __init__(self):
        super().__init__()
        self.fitted_count = 0

    def fit(self, X, y=None):
        self.fitted_count += 1
        return self

    def transform(self, X):
        return X + self.fitted_count

    def _more_tags(self):
        return {"requires_fit": True}


def _make_transformer(dir_name):
    pipeline = Pipeline(
        [
            (
                "preprocessor",
                wrap_bob_legacy(
                    FakePreprocessor(),
                    dir_name,
                    transform_extra_arguments=(("annotations", "annotations"),),
                ),
            ),
            (
                "extractor",
                wrap_bob_legacy(
                    FakeExtractor(),
                    dir_name,
                ),
            ),
            ("fittable_transformer", wrap(["sample"], FittableTransformer())),
        ]
    )

    return pipeline


def test_pipeline_train_function():
    with tempfile.TemporaryDirectory() as output:
        pipeline = _make_transformer(output)
        database = DummyDatabase()
        execute_pipeline_train(pipeline, database, output=output)
        assert os.path.isfile(os.path.join(output, "fittable_transformer.pkl"))


def _create_test_config_pipeline_simple(path):
    with open(path, "w") as f:
        f.write(
            """
from tests.test_pipeline_train import DummyDatabase, _make_transformer
from bob.bio.base.pipelines import PipelineSimple
from bob.bio.base.algorithm import Distance

database = DummyDatabase()

transformer = _make_transformer(".")

biometric_algorithm = Distance()

pipeline = PipelineSimple(
    transformer,
    biometric_algorithm,
    None,
)
"""
        )


def _create_test_config_pipeline_sklearn(path):
    with open(path, "w") as f:
        f.write(
            """
from tests.test_pipeline_train import DummyDatabase, _make_transformer

database = DummyDatabase()

pipeline = _make_transformer(".")
"""
        )


@pytest.mark.parametrize(
    "options,pipeline_simple",
    [
        (["--no-dask", "--memory"], True),
        (["--no-dask", "--memory"], False),
        (["--no-dask"], True),
        (["--no-dask"], False),
        (["--memory"], True),
        (["--memory"], False),
        ([], True),
        ([], False),
    ],
)
def test_pipeline_click_cli(
    options,
    pipeline_simple,
    expected_outputs=("results/fittable_transformer.pkl",),
):
    runner = CliRunner()
    with runner.isolated_filesystem():
        if pipeline_simple:
            _create_test_config_pipeline_simple("config.py")
        else:
            _create_test_config_pipeline_sklearn("config.py")
        result = runner.invoke(
            pipeline_train_cli,
            [
                "-vv",
                "config.py",
            ]
            + options,
        )
        assert_click_runner_result(result)
        # check for expected_output
        output_files = glob.glob("results/**", recursive=True)
        nl = "\n -"
        err_msg = f"Found only:\n- {nl.join(output_files)}\nin output directory given the options: {options}, and with {'PipelineSimple' if pipeline_simple else 'sklearn pipeline'}"
        for out in expected_outputs:
            assert os.path.isfile(out), err_msg



from . import Annotator


class Callable(Annotator):
    """A class that wraps a callable object that annotates a sample into a
    bob.bio.annotator object.

    Attributes
    ----------
    callable
        A callable with the following signature:
        ``annotations = callable(sample, **kwargs)`` that takes numpy array and
        returns annotations in dictionary format for that biometric sample.
        Please see :any:`Annotator` for more information.
    """

    def __init__(self, callable, **kwargs):
        super(Callable, self).__init__(**kwargs)
        self.callable = callable

    def transform(self, sample, **kwargs):
        return self.callable(sample, **kwargs)



from sklearn.base import BaseEstimator, TransformerMixin


class Annotator(TransformerMixin, BaseEstimator):
    """Annotator class for all annotators. This class is meant to be used in
    conjunction with the bob bio annotate script or to be used in pipelines.
    """

    def transform(self, samples, **kwargs):
        """Annotates a sample and returns annotations in a dictionary.

        Parameters
        ----------
        samples : numpy.ndarray
            The samples that are being annotated.
        **kwargs
            The extra arguments that may be passed.

        Returns
        -------
        dict
            A dictionary containing the annotations of the biometric sample. If
            the program fails to annotate the sample, it should return an empty
            dictionary.
        """
        raise NotImplementedError

    # Alias call to transform
    def __call__(self, samples, **kwargs):
        return self.transform(samples, **kwargs)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Manuel Guenther <Manuel.Guenther@idiap.ch>
# @date: Wed Oct 24 10:47:43 CEST 2012
#
# Copyright (C) 2011-2013 Idiap Research Institute, Martigny, Switzerland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This file defines a simple interface that are comparable with other bob.db databases.
"""

import fileinput
import os
import re


class FileListFile(object):
    """
    Initialize the File object with the minimum required data.

    If the ``model_id`` is not specified, ``model_id`` and ``client_id`` are identical.
    If the ``claimed_id`` is not specified, it is expected to be the ``client_id``.

    Parameters
    ----------

    client_id : various type
      The id of the client, this file belongs to.
      The type of it is dependent on your implementation.
      If you use an SQL database, this should be an SQL type like Integer or String.

    path : str
      The path of this file, relative to the basic directory.
      If you use an SQL database, this should be the SQL type String.
      Please do not specify any file extensions.

    file_id : various type
      The id of the file.
      The type of it is dependent on your implementation.
      If you use an SQL database, this should be an SQL type like Integer or String.
      If you are using an automatically determined file id, you can skip selecting the file id.
    """

    def __init__(self, file_name, client_id, model_id=None, claimed_id=None):
        # super(FileListFile, self).__init__(client_id=client_id, path=file_name, file_id=file_name)
        super(FileListFile, self).__init__()
        self.client_id = client_id
        self.path = file_name
        self.id = file_name

        # Note: in case of probe files, model ids are considered to be the ids of the model for the given probe file.
        # Hence, there might be several probe files with the same file id, but different model ids.
        # Therefore, please DO NOT USE the model_id outside of this class (or the according database queries).
        # when the model id is not specified, we use the client id instead
        self._model_id = client_id if model_id is None else model_id
        # when the claimed id is not specified, we use the client id instead
        self.claimed_id = client_id if claimed_id is None else claimed_id


#############################################################################
# internal access functions for the file lists; do not export!
#############################################################################


class ListReader(object):
    def __init__(self, store_lists):
        self.m_read_lists = {}
        self.m_model_dicts = {}
        self.m_store_lists = store_lists

    def _read_multi_column_list(self, list_file):
        rows = []
        if not os.path.isfile(list_file):
            raise RuntimeError("File %s does not exist." % (list_file,))
        try:
            for line in fileinput.input(list_file):
                if line.strip().startswith("#"):
                    continue
                parsed_line = re.findall(r"[\w/(-.)]+", line)
                if len(parsed_line):
                    # perform some sanity checks
                    if len(parsed_line) not in (2, 3, 4):
                        raise IOError(
                            "The read line '%s' from file '%s' could not be parsed successfully!"
                            % (line.rstrip(), list_file)
                        )
                    if len(rows) and len(rows[0]) != len(parsed_line):
                        raise IOError(
                            "The parsed line '%s' from file '%s' has a different number of elements than the first parsed line '%s'!"
                            % (parsed_line, list_file, rows[0])
                        )
                    # append the read line
                    rows.append(parsed_line)
            fileinput.close()
        except IOError as e:
            raise RuntimeError(
                "Error reading the file '%s' : '%s'." % (list_file, e)
            )

        # return the read list as a vector of columns
        return rows

    def _read_column_list(self, list_file, column_count):
        # read the list
        rows = self._read_multi_column_list(list_file)
        # extract the file from the first two columns
        file_list = []
        for row in rows:
            if column_count == 2:
                assert len(row) == 2
                # we expect: filename client_id
                file_list.append(
                    FileListFile(file_name=row[0], client_id=row[1])
                )
            elif column_count == 3:
                assert len(row) in (2, 3)
                # we expect: filename, model_id, client_id
                file_list.append(
                    FileListFile(
                        file_name=row[0],
                        client_id=row[2] if len(row) > 2 else row[1],
                        model_id=row[1],
                    )
                )
            elif column_count == 4:
                assert len(row) in (3, 4)
                # we expect: filename, model_id, claimed_id, client_id
                file_list.append(
                    FileListFile(
                        file_name=row[0],
                        client_id=row[3] if len(row) > 3 else row[1],
                        model_id=row[1],
                        claimed_id=row[2],
                    )
                )
            else:
                raise ValueError(
                    "The given column count %d cannot be interpreted. This is a BUG, please report to the author."
                    % column_count
                )

        return file_list

    def _create_model_dictionary(self, files):
        # remember model ids
        retval = {}
        for file in files:
            if file._model_id not in retval:
                retval[file._model_id] = file.client_id
            else:
                if retval[file._model_id] != file.client_id:
                    raise ValueError(
                        "The read model id '%s' is associated to two different client ids '%s' and '%s'!"
                        % (
                            file._model_id,
                            file.client_id,
                            retval[file._model_id],
                        )
                    )
        return retval

    def read_list(self, list_file, group, type=None):
        """Reads the list of Files from the given list file (if not done yet) and returns it."""
        if group in ("world", "optional_world_1", "optional_world_2"):
            if group not in self.m_read_lists:
                # read the world list into memory
                list = self._read_column_list(list_file, 2)
                if self.m_store_lists:
                    self.m_read_lists[group] = list
                return list
            # just return the previously read list
            return self.m_read_lists[group]

        else:
            if group not in self.m_read_lists:
                self.m_read_lists[group] = {}
            if type not in self.m_read_lists[group]:
                if type in ("for_models", "for_tnorm"):
                    list = self._read_column_list(list_file, 3)
                elif type == "for_scores":
                    list = self._read_column_list(list_file, 4)
                elif type in ("for_probes", "for_znorm"):
                    list = self._read_column_list(list_file, 2)
                else:
                    raise ValueError(
                        "The given type must be one of %s, but not '%s'"
                        % (
                            (
                                "for_models",
                                "for_scores",
                                "for_probes",
                                "for_tnorm",
                                "for_znorm",
                            ),
                            type,
                        )
                    )
                if self.m_store_lists:
                    self.m_read_lists[group][type] = list
                return list
            return self.m_read_lists[group][type]

    def read_models(self, list_file, group, type=None):
        """Generates a dictionary from model_ids to client_ids for the given list file, if not done yet, and returns it"""
        assert group in (
            "dev",
            "eval",
            "world",
            "optional_world_1",
            "optional_world_2",
        )
        assert type in ("for_models", "for_tnorm")
        if group not in self.m_model_dicts:
            self.m_model_dicts[group] = {}
        if type not in self.m_model_dicts[group]:
            dict = self._create_model_dictionary(
                self.read_list(list_file, group, type)
            )
            if self.m_store_lists:
                self.m_model_dicts[group][type] = dict
            return dict
        return self.m_model_dicts[group][type]
